import pytest
import fastpf


@pytest.fixture
def build_feeder3():
    # Tiny grid: Slack - PQ Node - Slack
    return fastpf.mockgrids.feeder(3, num_slacks=2)


@pytest.fixture
def build_meshed10():
    return fastpf.mockgrids.meshed(10)


# @pytest.mark.parametrize("pf_func", [fastpf.ybusnewton, fastpf.zbusjacobi])
# @pytest.mark.parametrize("pf_func", [fastpf.ybusnewton])
# def test_multiple_slacks_zero_load(pf_func, build_feeder3):
#     grid = build_feeder3
#     S = fastpf.mockloads.fixed(grid, load=0, num_loads=1)
#     U, iters, runtime = pf_func(grid, S)
#     assert U.size == len(grid["nodes"])
#     assert U[0, 0] == 400
#     assert abs(U[0, 1]) == 400
#     assert abs(U[0, 2]) == 401


# @pytest.mark.parametrize("pf_func", [fastpf.ybusnewton, fastpf.zbusjacobi])
# def test_multiple_slacks_low_load(pf_func, build_feeder3):
#     grid = fastpf.mockgrids.feeder(3, num_slacks=2)
#     S = fastpf.mockloads.fixed(grid, load=10, num_loads=1)
#     U, iters, runtime = pf_func(grid, S)
#     assert U.size == len(grid["nodes"])
#     assert U[0, 0] == 400
#     assert abs(U[0, 1]) < 400
#     assert abs(U[0, 2]) == 400


# @pytest.mark.parametrize("pf_func", [fastpf.ybusnewton, fastpf.zbusjacobi])
# def test_multiple_slacks_medium_load(pf_func):
#     grid = fastpf.mockgrids.feeder(3, num_slacks=2)
#     S = fastpf.mockloads.fixed(grid, load=10000, num_loads=1)
#     U, iters, runtime = pf_func(grid, S)
#     print(U)
#     assert U.size == len(grid["nodes"])
#     assert U[0, 0] == 400
#     assert abs(U[0, 1]) < 400
#     assert abs(U[0, 2]) == 400


# @pytest.mark.parametrize("pf_func", [fastpf.ybusnewton, fastpf.zbusjacobi])
# def test_multiple_slacks_10nodes(pf_func):
#     grid = fastpf.mockgrids.feeder(10, num_slacks=2)
#     S = fastpf.mockloads.fixed(grid, load=10000, num_loads=1)
#     U, iters, runtime = pf_func(grid, S)
#     print(U)
#     assert U.size == len(grid["nodes"])
#     assert U[0, 0] == 400
#     assert abs(U[0, 1]) < 400
#     assert abs(U[0, 9]) == 400



import pytest
import fastpf


@pytest.mark.parametrize("pf_func", [fastpf.ybusnewton, fastpf.zbusjacobi])
def test_multiple_slacks_zero_load(pf_func):
    grid = fastpf.testgrids.feeder(3, number_of_slacks=2)
    S = fastpf.testloads.fixed(grid, load_W=0, number_of_scenarios=1)
    U, all_converged, iters, runtime = pf_func(grid, S)
    assert all_converged
    assert U.size == len(grid["nodes"])
    assert U[0, 0] == 400
    assert abs(U[0, 1]) == 400
    assert abs(U[0, 2]) == 400


@pytest.mark.parametrize("pf_func", [fastpf.ybusnewton, fastpf.zbusjacobi])
def test_multiple_slacks_low_load(pf_func):
    grid = fastpf.testgrids.feeder(3, number_of_slacks=2)
    S = fastpf.testloads.fixed(grid, load_W=10, number_of_scenarios=1)
    U, all_converged, iters, runtime = pf_func(grid, S)
    assert all_converged
    assert U.size == len(grid["nodes"])
    assert U[0, 0] == 400
    assert abs(U[0, 1]) < 400
    assert abs(U[0, 2]) == 400


@pytest.mark.parametrize("pf_func", [fastpf.ybusnewton, fastpf.zbusjacobi])
def test_multiple_slacks_medium_load(pf_func):
    grid = fastpf.testgrids.feeder(3, number_of_slacks=2)
    S = fastpf.testloads.fixed(grid, load_W=10000, number_of_scenarios=1)
    U, all_converged, iters, runtime = pf_func(grid, S)
    assert all_converged
    assert U.size == len(grid["nodes"])
    assert U[0, 0] == 400
    assert abs(U[0, 1]) < 400
    assert abs(U[0, 2]) == 400


@pytest.mark.parametrize("pf_func", [fastpf.ybusnewton, fastpf.zbusjacobi])
def test_multiple_slacks_10nodes(pf_func):
    grid = fastpf.testgrids.feeder(10, number_of_slacks=2)
    S = fastpf.testloads.fixed(grid, load_W=10000, number_of_scenarios=1)
    U, all_converged, iters, runtime = pf_func(grid, S)
    assert all_converged
    assert U.size == len(grid["nodes"])
    assert U[0, 0] == 400
    assert abs(U[0, 1]) < 400
    assert abs(U[0, 9]) == 400



import pytest
import fastpf
import numpy as np


@pytest.mark.parametrize(
    "pf_func",
    [fastpf.ybusjacobi, fastpf.ybusgaussseidel, fastpf.ybusnewton, fastpf.zbusjacobi],
)
@pytest.mark.parametrize(
    "load_W",
    [0, 10, 10000],
)
def test_fixed_load_parallel(pf_func, load_W):
    grid = fastpf.testgrids.feeder(3)
    S = fastpf.testloads.fixed(grid, load_W=load_W, number_of_scenarios=1)

    U, all_converged, iters, runtime = pf_func(grid, S, num_processes=1)
    assert all_converged
    assert U.size == 3
    assert U[0, 0] == 400, f"with {pf_func.__name__}"

    U2, all_converged, iters, runtime = pf_func(grid, S, num_processes=2)
    assert all_converged
    assert U2.size == 3
    assert U2[0, 0] == 400, f"with {pf_func.__name__}"

    U4, all_converged, iters, runtime = pf_func(grid, S, num_processes=4)
    assert all_converged
    assert U4.size == 3
    assert U4[0, 0] == 400, f"with {pf_func.__name__}"

    assert np.allclose(U, U2, U4)



import numpy as np
import fastpf

grid = fastpf.testgrids.radial(30)
S = fastpf.testloads.random_uniform(grid, max_load_W=1000, number_of_scenarios=10000)

U_yjacobi, all_converged, iters, runtime = fastpf.ybusjacobi(grid, S)
U_yjacobi, all_converged, iters, runtime = fastpf.ybusjacobi(grid, S)
U_ygaussseidel, all_converged, iters, runtime = fastpf.ybusgaussseidel(grid, S)
U_newton, all_converged, iters, runtime = fastpf.ybusnewton(grid, S)
U_zbus, all_converged, iters, runtime = fastpf.zbusjacobi(grid, S)

print(np.allclose(U_newton, U_zbus, U_yjacobi, U_ygaussseidel))

fastpf.powerflow.compare_methods(grid, S)



import time
import numpy as np

# threadpoolctl allows control over the number of threads used by numpy,
# regardless of the underlying blas:
from threadpoolctl import threadpool_limits

# pathos allows multiprocessing with numpy arrays as parameters:
from pathos.multiprocessing import ProcessingPool

from . import powerflow_methods_cc as pf_cc
from . import process_grid
from . import integrate_slacks_for_Yident
from .log import logger
from .log import scale_time
from .log import set_loglevel

# This line replaces the compiled PF version with the Python versions, for development:
# from . import powerflow_methods as pf_cc


def _log_result(name, runtime_s, mean_iters, min_U, all_converged=True):
    runtime_string = scale_time(runtime_s)
    if all_converged:
        logger.success(
            f"{name:20s} | {runtime_string:>10} | {mean_iters:6.0f} | {min_U:.3f} V"
        )
    else:
        logger.warning(
            f"{name:20s} | {runtime_string:>10} | {mean_iters:6.0f} | {min_U:.3f} V - contains nonconverged results"
        )


def _parallelize(func, S, num_processes=1):
    # Shortcut for the case num_processes=1
    if num_processes == 1:
        U, iters = func(S)
        return U, iters

    if num_processes > 1:
        # Break up S into chunks:
        borders = np.linspace(0, S.shape[0], num_processes + 1, dtype=int)
        S_chunks = [S[borders[i] : borders[i + 1], :] for i in range(num_processes)]

        # Then process in parallel:
        with threadpool_limits(limits=1, user_api="blas"):
            with ProcessingPool(num_processes) as pool:
                all_results = pool.map(func, S_chunks)

        # And re-combine the results:
        U, iters = (result_set for result_set in zip(*all_results))
        U = np.vstack(U)
        iters = np.hstack(iters)

        return U, iters

    else:
        raise ValueError("num_processes must be > 1")


def ybusjacobi(grid, S, eps_s=1, max_iters=20000, num_processes=1):
    """Execute a power flow using the YBus Jacobi method.

    YBUS Jacobi has fast startup, is very memory-efficient and has very fast individual iterations, but potentially requires many iterations (>10000) to find a solution.

    ### Parameters:

    `grid`: *dict*, **required**
    > The grid dict (see ´fastpf.process_grid´) for which the power flow is executed

    `S`: *numpy array, complex 128*, **required**
    > The loads for which the power flow is executed, structured in a numpy array with *number_of_scenarios* rows (axis 0), and *number_of_nodes* columns (axis 1)

    `eps_s`: *int* or *float*, *Default: 1.0*
    > The accurracy at which an iterated result is accepted as final, in *VA*

    `max_iters`: *int*, *Default: 20000*
    > The number of iterations after which the iterations are interrupted. The intermediate, non-converged result is still stored in U, and the number of iterations stored in `iters` is max_iters + 1, so non-converged results can be checked with any(iters > max_iters)

    `num_processes`: *int*, *Default: 1*
    > Number of processes used for the computation. If >1, then `S` is chunked and recombined after the computation.

    ### Returns:
    `U`: *numpy array, complex128*
    > Resulting voltages

    `all_converged`: *Bool*
    > Is `False` if there are any non-converged results in U

    `iters`: *numpy array, int64*
    > Array with the iteration counts voltages

    `runtime_s`: *float*
    > Total runtime of the method in seconds
    """

    starttime = time.perf_counter()

    # Process grid and get relevant vars:
    grid_parameters = process_grid(grid, S)
    Y = grid_parameters["Y_ident"]
    U0 = grid_parameters["u0"]
    deleted_nodes = grid_parameters["node_ids_to_delete"]

    # Modify S to support slacks and deleted nodes:
    S = integrate_slacks_for_Yident(S, grid)
    S = np.delete(S, deleted_nodes, axis=1)

    # Perform the actual power flow in the compiled function:
    func = lambda S: pf_cc.ybusjacobi(Y, U0, S, eps_s, max_iters)
    U, iters = _parallelize(func, S, num_processes)

    # Check if all scenarios have converged:
    all_converged = all(iters <= max_iters)

    runtime_s = time.perf_counter() - starttime

    _log_result(
        "Ybus Jacobi", runtime_s, np.mean(iters), np.min(np.abs(U)), all_converged
    )

    return U, all_converged, iters, runtime_s


def ybusgaussseidel(
    grid,
    S,
    eps_s=1,
    max_iters=10000,
    acc_factor=1.6,
    num_processes=1,
):

    starttime = time.perf_counter()

    grid_parameters = process_grid(grid, S)
    Y = grid_parameters["Y_ident"]
    U0 = grid_parameters["u0"]

    S = integrate_slacks_for_Yident(S, grid)

    func = lambda S: pf_cc.ybusgaussseidel(Y, U0, S, acc_factor, eps_s, max_iters)
    U, iters = _parallelize(func, S, num_processes)

    all_converged = all(iters <= max_iters)

    runtime = time.perf_counter() - starttime

    _log_result(
        "Ybus Gauss Seidel", runtime, np.mean(iters), np.min(np.abs(U)), all_converged
    )

    return U, all_converged, iters, runtime


def _ybusrelaxation(grid, S, eps_s=1, max_iters=100000):
    starttime = time.perf_counter()
    grid_parameters = _parse_grid(grid)
    Y = grid_parameters["Y_ident"]
    U0 = grid_parameters["u0"]
    deleted_nodes = grid_parameters["node_ids_to_delete"]
    S = np.delete(S, deleted_nodes, axis=1)
    U, iters = pf_cc.ybusrelaxation(Y, U0, S, eps_s, max_iters)
    runtime = time.perf_counter() - starttime
    if verbose:
        _log_result("Ybus Relaxation", runtime, np.mean(iters), np.min(np.abs(U)))
    return U, iters, runtime


def ybusnewton(
    grid,
    S,
    eps_s=1,
    max_iters=100,
    num_processes=1,
    sparse=False,
):
    starttime = time.perf_counter()

    grid_parameters = process_grid(
        grid,
        S,
        reorder_slack_nodes=True,
    )
    Y = grid_parameters["Y"]
    Yred = grid_parameters["Yred"]
    U0 = grid_parameters["u0"]
    num_slacks = grid_parameters["numslacks"]

    logger.debug(f"Ybus Newton Setup Time: {time.perf_counter() - starttime:8.3f} s")
    if num_slacks > 1:
        additional_slack_voltages = grid_parameters["additional_slack_voltages"]
        m = len(additional_slack_voltages)
        for i, u_slack in enumerate(additional_slack_voltages, 1):
            U0[i] = u_slack
        func = lambda S: pf_cc.ybusnewton_mslacks(Y, S, U0, m, eps_s, max_iters)
    else:
        if sparse:
            slack_index = grid_parameters["slack_index"]
            deleted_nodes = grid_parameters["node_ids_to_delete"]
            S = np.delete(S, deleted_nodes, axis=1)

            func = lambda S: powerflow_methods.ybusnewton_sparse(
                Y, Yred, U0, S, slack_index, eps_s, max_iters
            )
        else:
            func = lambda S: pf_cc.ybusnewton(Y, S, U0, eps_s, max_iters)

    U, iters = _parallelize(func, S, num_processes)

    if grid_parameters["node_swaps"]:
        node_swaps = grid_parameters["node_swaps"].copy()
        while node_swaps:
            id1, id2 = node_swaps.pop()
            U[:, [id2, id1]] = U[:, [id1, id2]]

    all_converged = all(iters <= max_iters)

    runtime = time.perf_counter() - starttime

    _log_result(
        "Ybus Newton", runtime, np.mean(iters), np.min(np.abs(U)), all_converged
    )

    return U, all_converged, iters, runtime


def zbusjacobi(
    grid,
    S,
    U0=None,
    eps_s=1,
    max_iters=100,
    num_processes=1,
    sparse=False,
):
    starttime = time.perf_counter()

    grid_parameters = process_grid(grid, S)
    Zred = np.linalg.inv(grid_parameters["Yred"])
    if U0 is None:
        U0 = grid_parameters["u0"]
    slack_voltage = grid_parameters["main_slack_voltage"]
    deleted_nodes = grid_parameters["node_ids_to_delete"]
    slack_index = grid_parameters["slack_index"]
    S = grid_parameters["S"]
    additional_slack_ids = grid_parameters["additional_slack_ids"]
    S[:, additional_slack_ids] = 0

    logger.debug(f"Setup Time: {time.perf_counter() - starttime:8.3f} s")

    # if grid_parameters['node_swaps']:
    #     node_swaps = grid_parameters['node_swaps'].copy()
    #     print(node_swaps)
    #     while node_swaps:
    #         id1, id2 = node_swaps.pop()
    #         S[:,[id2, id1]] = S[:,[id1, id2]]

    # print(deleted_nodes, slack_index)

    S = np.delete(S, deleted_nodes + [slack_index], axis=1)

    ## 2 Slacks:
    if grid_parameters["numslacks"] == 2:
        i_slack = (
            additional_slack_ids[0] - 1
        )  # -1 because the slack index 0 is deleted in Zred and S
        U_slack2 = grid_parameters["additional_slack_voltages"][0]
        eps_u_slack = 0.001
        U, iters = pf_cc.zbusjacobi_2slacks(
            Zred, S, U0, slack_voltage, i_slack, U_slack2, eps_s, eps_u_slack, max_iters
        )

    ## n Slacks:
    elif grid_parameters["numslacks"] > 2:
        m = len(grid_parameters["additional_slack_voltages"])
        U_additional_slacks = grid_parameters["additional_slack_voltages"]
        eps_u_slack = 0.001
        par_func = lambda S: pf_cc.zbusjacobi_mslacks(
            Zred,
            S,
            U0,
            slack_voltage,
            m,
            U_additional_slacks,
            eps_s,
            eps_u_slack,
            max_iters,
        )
        U, iters = _parallelize(par_func, S, num_processes)
    # 1 slack:
    else:
        if sparse:
            func = lambda S: powerflow_methods.zbusjacobi_sparse(
                Zred, S, slack_voltage, eps_s, max_iters
            )
        else:
            func = lambda S: pf_cc.zbusjacobi(
                Zred, S, U0, slack_voltage, eps_s, max_iters
            )
        U, iters = _parallelize(func, S, num_processes)

    # If slack nodes have been swapped to the top, swap them back so that the original order is restored:
    if grid_parameters["node_swaps"]:
        node_swaps = grid_parameters["node_swaps"].copy()
        while node_swaps:
            id1, id2 = node_swaps.pop()
            U[:, [id2, id1]] = U[:, [id1, id2]]

    all_converged = all(iters <= max_iters)

    runtime = time.perf_counter() - starttime

    _log_result(
        "Zbus Jacobi", runtime, np.mean(iters), np.min(np.abs(U)), all_converged
    )

    return U, all_converged, iters, runtime


def _bfs(grid, S, eps_s=1, max_iters=1000):
    starttime = time.perf_counter()
    grid_parameters = _parse_grid(grid)
    Zline = grid_parameters["Zline"]
    if Zline is None:
        name = "& Backward/Forward Sweep"
        print(f"{name:25s} & - & - \\\\")
        # print('Error: BFS is impossible because grid is not a feeder')
        return 0, 0, 0
    slack_voltage = grid_parameters["u0"][0]
    U, iters = pf_cc.bfs(Zline, S, slack_voltage, eps_s, max_iters)
    runtime = time.perf_counter() - starttime
    if verbose:
        _log_result(
            "Backward/Forward Sweep", runtime, np.mean(iters), np.min(np.abs(U))
        )
    return U, iters, runtime


def compare_methods(grid, S):
    from rich.console import Console
    from rich.table import Table

    console = Console()

    table = Table(show_header=True, header_style="bold green")
    table.add_column("Method")
    table.add_column("Iters", justify="right")
    table.add_column("Runtime", justify="right")

    pf_funcs = (
        ybusjacobi,
        ybusgaussseidel,
        ybusnewton,
        zbusjacobi,
    )

    set_loglevel("ERROR")

    with console.status("Computing power flow...") as status:
        for func in pf_funcs:
            U, all_converged, iters, runtime = func(grid, S)
            console.log(f"{func.__name__} done")
            table.add_row(func.__name__, str(np.mean(iters)), scale_time(runtime))

    console.print(table)



import pytest
import fastpf


@pytest.fixture
def build_feeder3():
    return fastpf.testgrids.feeder(3)


@pytest.fixture
def build_meshed10():
    return fastpf.testgrids.meshed(10)


@pytest.mark.parametrize(
    "pf_func",
    [fastpf.ybusjacobi, fastpf.ybusgaussseidel, fastpf.ybusnewton, fastpf.zbusjacobi],
)
def test_zero_load(pf_func):
    grid = fastpf.testgrids.feeder(3)
    S = fastpf.testloads.fixed(grid, load_W=0, number_of_scenarios=1)
    U, all_converged, iters, runtime = pf_func(grid, S)
    assert U.size == 3
    assert U[0, 0] == 400, f"with {pf_func.__name__}"
    assert abs(U[0, 1]) == 400, f"with {pf_func.__name__}"
    assert abs(U[0, 2]) == 400, f"with {pf_func.__name__}"


@pytest.mark.parametrize(
    "pf_func",
    [fastpf.ybusjacobi, fastpf.ybusgaussseidel, fastpf.ybusnewton, fastpf.zbusjacobi],
)
def test_low_load(pf_func, build_feeder3):
    grid = build_feeder3
    S = fastpf.testloads.fixed(grid, load_W=10, number_of_scenarios=1)
    U, all_converged, iters, runtime = pf_func(grid, S)
    assert U.size == 3
    assert U[0, 0] == 400
    assert abs(U[0, 1]) > 399.9, f"with {pf_func.__name__}"
    assert abs(U[0, 1]) < 400, f"with {pf_func.__name__}"
    assert abs(U[0, 2]) > 399.9, f"with {pf_func.__name__}"
    assert abs(U[0, 2]) < 400, f"with {pf_func.__name__}"


@pytest.mark.parametrize(
    "pf_func",
    [fastpf.ybusjacobi, fastpf.ybusgaussseidel, fastpf.ybusnewton, fastpf.zbusjacobi],
)
def test_medium_load(pf_func, build_feeder3):
    grid = build_feeder3
    S = fastpf.testloads.fixed(grid, load_W=10000, number_of_scenarios=1)
    U, all_converged, iters, runtime = pf_func(grid, S)
    assert U.size == 3, f"with {pf_func.__name__}"
    assert U[0, 0] == 400, f"with {pf_func.__name__}"
    assert abs(U[0, 1]) > abs(U[0, 2]), f"with {pf_func.__name__}"

    # assert abs(U[0, 1]) > 399.77, f"with {pf_func.__name__}"
    # assert abs(U[0, 1]) < 399.78, f"with {pf_func.__name__}"
    # assert abs(U[0, 2]) > 399.66, f"with {pf_func.__name__}"
    # assert abs(U[0, 2]) < 399.67, f"with {pf_func.__name__}"


@pytest.mark.parametrize(
    "pf_func",
    [fastpf.ybusjacobi, fastpf.ybusgaussseidel, fastpf.ybusnewton, fastpf.zbusjacobi],
)
def test_meshed(pf_func):
    grid = fastpf.testgrids.meshed(10)
    S = fastpf.testloads.fixed(grid, load_W=1000, number_of_scenarios=1)

    U, all_converged, iters, runtime = fastpf.ybusjacobi(grid, S)
    assert all_converged
    assert U.size == 10
    assert U[0, 0] == 400
    for i in range(1, 10):
        assert abs(U[0, i]) < 400
        assert abs(U[0, i]) > 399


# def test_allclose():
#     grid = fastpf.testgrids.feeder(10)
#     S = fastpf.testestds.random(grid, maxload=1000, number_of_loads=1000)



import os
import copy
import networkx as nx

from matplotlib import rc
import matplotlib.font_manager

rc("text", usetex=True)
rc("font", family="serif")
# required for an upright, non-italic mu:
rc("text.latex", preamble=r"\usepackage{upgreek}")

# This allows import fastpf.plotting as plt:
from matplotlib.pyplot import *

# TODO: potentally not zip-safe, needs to be done with pkg_resources:
style.use(os.path.join(os.path.dirname(__file__), "diss.mplstyle"))

cmap = get_cmap("Blues")
cmap_greens = get_cmap("Greens")

# Bold font:
boldfont = {"fontname": "CMU Serif"}

# Gray box for legends and text boxes:
graybox = {"boxstyle": "round", "linewidth": 0, "facecolor": "lightgray"}

# Helper Function so you can set size as a multiplicator of 2.5 inches with plt.setsize(fig,1...n)
def setsize(fig, size=1):
    fig.set_size_inches((4.33, 2.5 * size))


# Helper Function to reach the images folder:
to_images = functools.partial(
    os.path.join, os.getenv("HOME"), "Dropbox", "diss", "images"
)


def save(fig, filename):
    if "." not in filename:
        filename += ".eps"
    fig.savefig(to_images(filename), dpi=600, bbox_inches="tight", pad_inches=0.04)
    print(f"Figure saved to {filename}")


def plotgraph(
    grid,
    filename="",
    S=None,
    I=None,
    ax=None,
    shape=None,
    node_labels=False,
    lines_removed=[],
    verbose=False,
    **additional_args,
):
    grid = copy.deepcopy(grid)
    g = nx.OrderedGraph()
    nodes = grid["nodes"]
    node_ids = [g["id"] for g in nodes]
    if verbose:
        print(f"Plotting {len(node_ids)} Nodes...")

    g.add_nodes_from(node_ids)
    edge_color = []
    edgelist = []
    for line in sorted(grid["edges"], key=lambda line: line["id"]):
        g.add_edge(line["source"], line["target"])
        # Set Line Colors:
        edgelist.append(
            (line["source"], line["target"])
        )  # Explicit edgelist is required so that edge_color works properly
        if I is not None:
            Imax = np.max(np.abs(I))
            cmap = matplotlib.cm.get_cmap("OrRd")
            edge_color.append(cmap(np.max(np.abs(I[:, line["id"]])) / Imax))
            if line["id"] in lines_removed:
                edge_color[-1] = "blue"
        else:
            edge_color = "gray"

    # Set 'noload' for Node Colors and sizes:
    # =======================================
    if S is not None:
        for node in nodes:
            if np.sum(np.abs(S[:, node["id"]])) != 0:
                node["noload"] = False
            else:
                node["noload"] = True
    else:  # If S is not provided, noload=False for every node:
        for i, n in enumerate(nodes):
            nodes[i]["noload"] = False

    # Set Node Positions:
    # ===================
    if shape == "spring":
        positions = nx.spring_layout(g)
    elif shape == "force":
        starting_positions = None
        # if 'x' in nodes[0] and 'y' in nodes[0]:
        #     starting_positions = {n['id']:(n['x'], n['y']) for n in nodes}
        positions = nx.kamada_kawai_layout(g, pos=starting_positions)
    elif shape is None:
        if "x" in nodes[0] and "y" in nodes[0]:
            positions = {n["id"]: (n["x"], n["y"]) for n in nodes}
        elif "xy" in nodes[0]:
            positions = {n["id"]: (n["xy"][0], n["xy"][1]) for n in nodes}
        else:
            positions = nx.kamada_kawai_layout(g)

    # Set Slacks to green, loads to orange, rest to gray:
    # Filter nodes with PV and color blue:
    for i, node in enumerate(nodes):
        if "elements" in node:
            connected_element_types = [el["type"] for el in node["elements"]]
        else:
            connected_element_types = []
        if "PV" in connected_element_types:
            nodes[i]["has_pv"] = True
        else:
            nodes[i]["has_pv"] = False

    if "color" in nodes[0]:
        node_color = [n["color"] for n in nodes]
    else:
        node_color = [
            "orangered" if not n["noload"] else "dimgray" for n in grid["nodes"]
        ]
        node_color = [
            "blue" if n["has_pv"] else nc for n, nc in zip(grid["nodes"], node_color)
        ]
        node_color = [
            "green" if n["is_slack"] else nc for n, nc in zip(grid["nodes"], node_color)
        ]

    if "size" in nodes[0]:
        node_size = [n["size"] for n in nodes]
    else:
        node_size = [18 if not n["noload"] else 3 for n in grid["nodes"]]
        node_size = [
            18 if n["is_slack"] else ns for n, ns in zip(grid["nodes"], node_size)
        ]

    if ax == None:
        fig, ax = subplots()
        setsize(fig, 1)
        ax.axis("off")
        margins(0.01)
        ax.margins(0.02)
        showplot = True
    else:
        fig = None
        ax.axis("off")
        showplot = False

    nx.draw_networkx(
        g,
        ax=ax,
        pos=positions,
        node_color=node_color,
        node_size=node_size,
        edgelist=edgelist,
        edge_color=edge_color,
        width=1.5,
        with_labels=node_labels,
        **additional_args,
    )

    tight_layout()

    if filename:
        if not fig:
            print("Cant save plot in external axes")
        else:
            save(fig, f"{filename}.eps")

    if showplot:
        show()


def plot_result(grid, S, U, I, lines_removed=[], verbose=False):

    if verbose:
        for line in grid["edges"]:
            print(
                f'Line {line["id"]:>3}:   {line["source"]:>3} - {line["target"]:>3}:    R,X = {line["R"]:7.3f}, {line["X"]:7.3f}    I = {abs(I)[:,line["id"]].max():.2f}'
            )

    fig = figure()
    gs = matplotlib.gridspec.GridSpec(3, 2, figure=fig)

    ax0 = fig.add_subplot(gs[0, 0])
    ax0.set_title("Abs. Power S / VA")
    plot(np.sign(S.real) * np.abs(S))
    ax0.grid(True)

    ax1 = fig.add_subplot(gs[1, 0])
    ax1.set_title("Abs. Voltage / V")
    plot(np.abs(U), alpha=0.5)
    ax1.grid(True)

    ax2 = fig.add_subplot(gs[2, 0])
    bar(range(I.shape[1]), np.max(np.abs(I), axis=0), zorder=2)
    # plt.plot(np.sign(I.real) * np.abs(I))
    ax2.grid(True)

    ax3 = fig.add_subplot(gs[:, 1])
    plotgraph(grid, I=I, S=S, ax=ax3, node_labels=False, lines_removed=lines_removed)

    tight_layout()
    show()



import sys
from loguru import logger


def scale_time(time_s):
    """
    Transform a time in seconds to a string with the appropriate unit
    """
    if time_s >= 60:  # >= 1 minute
        minutes = round(time_s // 60)
        seconds = round(time_s % 60)
        return f"{minutes}min {seconds}s"
    if time_s >= 1:  # >= 1 second
        seconds = round(time_s % 60)
        milliseconds = round((time_s % 1) * 1000)
        return f"{seconds}.{milliseconds}s"
    if time_s < 1:  # < 1 second
        milliseconds = round(time_s * 1000)
        return f"{milliseconds}ms"


def set_loglevel(loglevel):

    if loglevel.upper() not in ["DEBUG", "INFO", "ERROR", "OFF"]:
        raise ValueError(
            f"loglevel must be 'DEBUG', 'INFO', 'ERROR', or 'OFF', not {loglevel}"
        )

    logger.remove()

    if loglevel.upper() == "DEBUG":
        logger.add(
            sink=sys.stdout,
            level="DEBUG",
            format="{time:HH:mm:ss} | {elapsed} | <cyan>{module}:{function}</cyan> | <lvl>{message}</lvl>",
        )

    if loglevel.upper() == "INFO":
        logger.add(
            sink=sys.stdout,
            level="INFO",
            format="{time:HH:mm:ss} | {elapsed} | <lvl>{message}</lvl>",
        )

    if loglevel.upper() == "ERROR":
        logger.add(
            sink=sys.stdout,
            level="ERROR",
            format="{time:HH:mm:ss} | {elapsed} | <lvl>{message}</lvl>",
        )

    if loglevel.upper() == "OFF":
        pass


# set_loglevel("INFO")



import matplotlib.pyplot as plt


def plot_phasor_points(U):
    plt.plot(U.real, U.imag, ".")
    plt.savefig("U.png")


def plot_phasor_points_radial(U, save_as=None):
    plt.plot(np.abs(U), np.angle(U), ".")
    if save_as:
        plt.savefig(f"{save_as}.png")



import time
import copy
import collections
import numpy as np

from . import validation
from .log import logger


def extract_zline(grid):
    # if numberofnodes != numberoflines + 1:
    #     return None

    # if not all([c == 1 for c in sourcenodes_count]):
    #     return None

    sourcenodes = [l["source"] for l in grid["edges"]]
    sourcenodes_count = collections.Counter(sourcenodes).values()

    Zline = np.zeros(len(sourcenodes), dtype=np.complex128)

    for i, line in enumerate(sorted(grid["edges"], key=lambda l: l["source"])):
        Zline[i] = line["R"] + 1j * line["X"]

    return Zline


def integrate_slacks_into_S(S, grid):
    S_copy = S.copy()
    for i, node in enumerate(grid["nodes"]):
        if node["is_slack"]:
            S_copy[:, i] = (
                np.abs(
                    node["slack_voltage"]
                    * np.exp(1j * node["slack_angle"] / 360 * 2 * np.pi)
                )
                ** 2
            )
    return S_copy


# def extract_U0(grid):
#     number_of_nodes = len(grid["nodes"])
#     U0 = np.zeros((1, number_of_nodes), dtype=np.complex128)
#     for n in grid["nodes"]:
#         u0 = 0 + 0j
#         if "P" in n:
#             S += float(P)
#         if "Q" in n:
#             S += 1j * float(Q)
#     return U0


def extract_S(grid):
    number_of_nodes = len(grid["nodes"])
    S = np.zeros((1, number_of_nodes), dtype=np.complex128)
    for n in grid["nodes"]:
        S[0, n["id"]] = 0 + 0j
        if "P" in n:
            S[0, n["id"]] += float(n["P"])
        if "Q" in n:
            S[0, n["id"]] += 1j * float(n["Q"])
    return S


def collapse_slacks(grid, main_slack_id=None):
    logger.debug("Collapsing identical slack nodes...")

    # Collect Slack Nodes:
    slacknodes = [n for n in grid["nodes"] if n["is_slack"]]
    number_of_slacks = len(slacknodes)

    # If there is only 1 slack, there is nothing to do:
    if number_of_slacks == 1:
        logger.debug(" -> grid has only one slack node, grid unchanged")
        return grid, []

    # If main_slack_id is not explicitly given, its the node with the lowest id:
    if main_slack_id is None:
        main_slack_id = slacknodes[0]["id"]

    # Deep copy before any changes are made:
    grid = copy.deepcopy(grid)
    nodes = grid["nodes"]
    lines = grid["lines"]

    additional_slack_ids = {n["id"] for n in slacknodes} - {main_slack_id}

    deleted_node_ids = []
    for slack1 in [main_slack_id] + list(additional_slack_ids):
        for slack2 in additional_slack_ids:
            if slack2 != slack1 and slack1:
                same_voltages = (
                    nodes[slack1]["slack_voltage"] == nodes[slack2]["slack_voltage"]
                )
                same_angles = (
                    nodes[slack1]["slack_angle"] == nodes[slack2]["slack_angle"]
                )
                if same_voltages and same_voltages:
                    for line in lines:
                        if line["from"] == slack2:
                            line["from"] = slack1
                            logger.debug(
                                f' -> Line {line["id"]} "from"-node: {slack2} -> {slack1}'
                            )
                        if line["to"] == slack2:
                            line["to"] = slack1
                            logger.debug(
                                f' -> Line {line["id"]} "to"-node: {slack2} -> {slack1}'
                            )
                    deleted_node_ids.append(slack2)
                    logger.debug(
                        f" -> Deleting slack node {slack2} and integrating into {slack1}"
                    )
    nodes = [node for node in nodes if node["id"] not in deleted_node_ids]
    grid["nodes"] = nodes

    return grid, deleted_node_ids


def reorder_nodes(grid, S, main_slack_id=None):
    logger.debug(f"Reordering Nodes to bring slacks to top...")

    # Collect slacks:
    slack_nodes = [n for n in grid["nodes"] if n["is_slack"]]
    number_of_slacks = len(slack_nodes)

    # If main_slack_id is not explicitly given, its the node with the lowest id:
    if main_slack_id is None:
        main_slack_id = slack_nodes[0]["id"]
    additional_slack_ids = {n["id"] for n in slack_nodes} - {main_slack_id}

    # 1 slack node in ID 0 -> correct -> return:
    if number_of_slacks == 1 and grid["nodes"][0]["is_slack"] and main_slack_id == 0:
        logger.debug(" -> only slack node is in ID 0, grid unchanged")
        return grid, S, []

    # Check if multiple slacks are correct -> return:
    # slack_ids_correct = True
    # if main_slack_id == 0 and grid["nodes"][0]["is_slack"]:
    #     for slack_node in range(number_of_slacks):
    #         if not grid["nodes"][slack_node]["is_slack"]:
    #             slack_ids_correct = False
    # if slack_ids_correct:
    #     logger.debug(
    #         f" -> {number_of_slacks} slack nodes are in correct order, grid unchanged"
    #     )
    #     return grid, S, []

    # Deep copy before any changes are made:
    grid = copy.deepcopy(grid)
    nodes = grid["nodes"]
    lines = grid["lines"]

    # Sort slack to tops according to this list:
    slack_list = [main_slack_id] + list(additional_slack_ids)

    node_swaps = []
    # Walk through list and swap nodes:
    for i, node_id in enumerate(slack_list):
        if node_id is not i:
            logger.debug(f" -> Swapping nodes {node_id} and {i}")
            for line in lines:
                if line["from"] == i:
                    line["from"] = node_id
                elif line["from"] == node_id:
                    line["from"] = i
                if line["to"] == i:
                    line["to"] = node_id
                elif line["to"] == node_id:
                    line["to"] = i
            for node in nodes:
                if node["id"] == i:
                    node1 = node
                elif node["id"] == node_id:
                    node2 = node
            node1["id"] = node_id
            node2["id"] = i
            S[:, [node_id, i]] = S[:, [i, node_id]]
            node_swaps.append((i, node_id))
    logger.debug(f" -> Slack IDs after reordering: 0 + {additional_slack_ids}")

    nodes.sort(key=lambda n: n["id"])
    grid["nodes"] = nodes

    return grid, S, node_swaps


def process_grid(
    grid,
    S=None,
    main_slack_id=None,
    collapse_identical_slacks=False,
    reorder_slack_nodes=True,
):

    logger.debug("Processing grid...")

    grid = copy.deepcopy(grid)

    if S is None:
        logger.debug(" -> Extracting S from grid dict...")
        S = extract_S(grid)

    # Check Validity of grid and S:
    errors = validation.validate_grid(grid, S)
    if errors:
        raise Exception(f"{len(errors)} errors found in grid during validation")

    # Collect Slack Nodes:
    slacknodes = [n for n in grid["nodes"] if n["is_slack"]]
    numberofslacks = len(slacknodes)

    if numberofslacks == 1:
        main_slack_id = slacknodes[0]["id"]
        additional_slack_ids = {}
    else:
        if not main_slack_id:
            main_slack_id = slacknodes[0]["id"]

    # Delete and reconnect nodes with identical slack voltage:
    if collapse_identical_slacks:
        grid, deleted_node_ids = collapse_slacks(grid, main_slack_id=main_slack_id)
    else:
        deleted_node_ids = []

    # Reorder to bring slacks to top:
    grid, S, node_swaps = reorder_nodes(grid, S, main_slack_id=main_slack_id)

    # At this point, nodes and lines are fixed:
    nodes = grid["nodes"]
    lines = grid["lines"]
    numberofnodes = len(nodes)
    numberoflines = len(lines)
    slacknodes = [n for n in nodes if n["is_slack"]]
    main_slack_id = 0
    main_slack_voltage = nodes[main_slack_id]["slack_voltage"]
    main_slack_angle = nodes[main_slack_id]["slack_angle"]
    additional_slack_ids = {n["id"] for n in slacknodes} - {main_slack_id}

    logger.debug(f"Number of Nodes: {numberofnodes}")
    logger.debug(f"Number of Lines: {numberoflines}")
    logger.debug(f"Number of Slacks: {numberofslacks}")

    # Construct the admittance matrix Y:
    # ==================================

    # Preallocate admittance matrix:
    Y = np.zeros((numberofnodes, numberofnodes), dtype=np.complex128)

    # Add the connections in 'lines' to the admittance matrix Y
    for i, line in enumerate(lines):
        # Make sure connection goes from lower node number to higher node number,
        # for easy symmetrization with np.transpose later:
        if line["from"] > line["to"]:
            from_node, to_node = line["to"], line["from"]
        else:
            from_node, to_node = line["from"], line["to"]

        # Make the connection in the admittancy matrix:
        Y[from_node, to_node] += 1 / (line["R"] + 1j * line["X"])

    # Symmetrize:
    Y = Y + np.transpose(Y)

    # Assign diagonal elements:
    for i in range(0, len(Y)):
        Y[i, i] = -np.sum(Y[i, :])

    # Process shunt impedances:
    # for node in nodes:
    #     i = node["id"]
    #     if "B" not in node:
    #         node["B"] = 0
    #     if "G" not in node:
    #         node["G"] = 0
    #     Y[i, i] = node["G"] + 1j * node["B"]

    # ==================================

    additional_slack_voltages = np.array(
        [
            nodes[i]["slack_voltage"]
            * np.exp(1j * nodes[i]["slack_angle"] / 360 * 2 * np.pi)
            for i in additional_slack_ids
        ],
        dtype=np.complex128,
    )

    # Create Y_ident which has ident line for slack nodes:
    Y_ident = Y.copy()
    all_slack_ids = [main_slack_id] + list(additional_slack_ids)
    Y_ident[all_slack_ids, :] = np.zeros_like(Y_ident[all_slack_ids, :])
    Y_ident[all_slack_ids, all_slack_ids] = 1

    # Prepare u0:
    u0 = np.ones(numberofnodes, dtype=np.complex128) * main_slack_voltage
    u0[main_slack_id] = nodes[main_slack_id]["slack_voltage"] * np.exp(
        1j * nodes[main_slack_id]["slack_angle"] / 360 * 2 * np.pi
    )

    # Calc Yred:
    Yred = np.delete(Y, main_slack_id, axis=0)
    Yred = np.delete(Yred, main_slack_id, axis=1)

    # Check if Yred is invertible and change line to ident if not:
    # This 'neutralizes' nodes that would be isolated because of a deleted line etc.
    # if np.linalg.det(Yred) == 0:
    #     for n in range(Yred.shape[0]):
    #         if np.sum(np.abs(Yred[n,:])) == 0:
    #             logger.debug(f'Neutralizing Node {n}')
    #             Yred[n,n] = 1
    #             S[:,n] = 0

    # If grid is a feeder, construct Zline for BFS:
    # is_feeder, is_radial = determine_gridshape(grid)
    # if is_feeder:
    #     Z_line = extract_zline(grid)
    # else:
    #     Z_line = None

    grid["nodes"] = nodes
    grid["edges"] = lines

    grid_parameters = {
        "transformed_grid": grid,
        "Y": Y,
        "Yred": Yred,
        "Y_ident": Y_ident,
        "u0": u0,
        "original_slack_indices": [n["id"] for n in slacknodes],
        "slack_index": main_slack_id,
        "main_slack_voltage": nodes[main_slack_id]["slack_voltage"]
        * np.exp(1j * nodes[main_slack_id]["slack_angle"] / 360 * 2 * np.pi),
        "node_ids_to_delete": deleted_node_ids,
        "original_number_of_slacks": len(slacknodes),
        "numslacks": numberofslacks,
        "additional_slack_ids": list(additional_slack_ids),
        "additional_slack_voltages": additional_slack_voltages,
        "node_swaps": node_swaps,
        "S": S,
    }

    return grid_parameters


def reorder_grid_nodes(
    grid, S=None, collapse_identical_slacks=False, main_slack_id=None, verbose=0
):

    if verbose > 0:
        log = print
    else:
        log = lambda *args, **kwargs: None

    nodes = grid["nodes"]
    lines = grid["edges"]
    slacknodes = [n for n in nodes if n["is_slack"]]

    if len(slacknodes) == 1:
        main_slack_id = slacknodes[0]["id"]
        additional_slack_ids = {}
    else:
        if not main_slack_id:
            main_slack_id = slacknodes[0]["id"]
        additional_slack_ids = {n["id"] for n in slacknodes} - {main_slack_id}

    # Delete nodes with identical slack voltage and reconnect:
    # ========================================================
    node_ids_to_delete = []
    if collapse_identical_slacks:
        for slack1 in [main_slack_id] + list(additional_slack_ids):
            for slack2 in additional_slack_ids:
                if slack2 != slack1 and slack1 not in node_ids_to_delete:
                    same_voltages = (
                        nodes[slack1]["slack_voltage"] == nodes[slack2]["slack_voltage"]
                    )
                    same_angles = (
                        nodes[slack1]["slack_angle"] == nodes[slack2]["slack_angle"]
                    )
                    if same_voltages and same_voltages:
                        for line in lines:
                            if line["from"] == slack2:
                                line["from"] = slack1
                                log(
                                    f'Redirecting line {line["id"]} from node: {slack2} -> {slack1}'
                                )
                            if line["to"] == slack2:
                                line["to"] = slack1
                                log(
                                    f'Redirecting line {line["id"]} to node: {slack2} -> {slack1}'
                                )
                        node_ids_to_delete.append(slack2)
                        log(
                            f"Deleting slack node {slack2} and integrating into {slack1}"
                        )

        # Deleted nodes from the 'nodes' list:
        nodes = [node for node in nodes if node["id"] not in node_ids_to_delete]
        S = np.delete(S, node_ids_to_delete, axis=1)

    # Normalize node IDs & fill the gaps:
    nn = []  # normalized nodes
    node_id_dict = {n["id"]: n for n in nodes}
    for i in range(len(nodes)):
        original_id = i
        if i in node_id_dict:
            nn.append(node_id_dict[i])
        else:
            next_node_found = False
            while not next_node_found:
                i += 1
                if i in node_id_dict:
                    print(original_id, i)
                    next_node = node_id_dict[i]
                    del node_id_dict[i]
                    next_node["id"] = original_id
                    nn.append(next_node)
                    for l in lines:
                        if l["from"] == i:
                            l["from"] = original_id
                        if l["to"] == i:
                            l["to"] = original_id
                    next_node_found = True

        # node_id_dict = {n['id']:n for n in nodes}
    print([n["id"] for n in nn])

    nodes = nn
    slacknodes = [n for n in nodes if n["is_slack"]]

    if len(slacknodes) == 1:
        main_slack_id = slacknodes[0]["id"]
        additional_slack_ids = {}
    else:
        if not main_slack_id:
            main_slack_id = slacknodes[0]["id"]
        additional_slack_ids = {n["id"] for n in slacknodes} - {main_slack_id}

    # Reorder to bring slacks to top:
    # ===============================
    # plt.plotgraph(grid, node_labels=True)
    node_swaps = []
    new_additional_slack_ids = []
    for i, node_id in enumerate([main_slack_id] + list(additional_slack_ids)):
        if node_id is not i:
            log(f"Swapping nodes {node_id} and {i}")
            for line in lines:
                if line["from"] == i:
                    line["from"] = node_id
                elif line["from"] == node_id:
                    line["from"] = i
                if line["to"] == i:
                    line["to"] = node_id
                elif line["to"] == node_id:
                    line["to"] = i
            for node in nodes:
                if node["id"] == i:
                    node1 = node
                elif node["id"] == node_id:
                    node2 = node
            node1["id"] = node_id
            node2["id"] = i
            if S is not None:
                S[:, [node_id, i]] = S[:, [i, node_id]]
            node_swaps.append((i, node_id))
        new_additional_slack_ids.append(i)
    main_slack_id = 0
    additional_slack_ids = set(new_additional_slack_ids) - {0}
    log(f"Slack IDs after reordering: 0 + {additional_slack_ids}")
    nodes.sort(key=lambda n: n["id"])

    grid["nodes"] = nodes
    grid["edges"] = lines

    return grid, node_swaps, S


def calcY(grid, lines_to_remove=[]):
    lines = grid["edges"]
    num_nodes = len(grid["nodes"])
    Y = np.zeros((num_nodes, num_nodes), dtype=np.complex128)
    # Add the connections in 'lines' to the admittancy matrix Y
    for i, line in enumerate(lines):
        if line["id"] not in lines_to_remove:
            Y[line["to"], line["from"]] += 1 / (line["R"] + 1j * line["X"])
            Y[line["from"], line["to"]] += 1 / (line["R"] + 1j * line["X"])

    # Assign diagonal elements:
    for i in range(Y.shape[0]):
        Y[i, i] = -np.sum(Y[i, :])

    return Y


def put_slack_voltages_into_u0(grid, base_voltage, u0=None):
    num_nodes = len(grid["nodes"])
    u0 = np.zeros(num_nodes, dtype=np.complex128)
    for node in grid["nodes"]:
        if node["is_slack"]:
            u0[node["id"]] = node["slack_voltage"] * np.exp(
                1j * node["slack_angle"] / 360 * 2 * np.pi
            )
        else:
            u0[node["id"]] = base_voltage
    return u0


def calc_line_impedances(grid, lines_to_remove=[]):
    line_impedances = np.zeros(
        (len(grid["edges"]), len(grid["nodes"])), dtype=np.complex128
    )

    for line in grid["edges"]:
        if line["id"] not in lines_to_remove:
            line_impedances[line["id"], line["from"]] += 1 / (
                line["R"] + 1j * line["X"]
            )
            line_impedances[line["id"], line["to"]] -= 1 / (line["R"] + 1j * line["X"])

    return line_impedances


def integrate_slacks_for_Yident(S, grid):
    S = S.copy()
    for i, node in enumerate(grid["nodes"]):
        if node["is_slack"]:
            U_0 = np.abs(
                node["slack_voltage"]
                * np.exp(1j * node["slack_angle"] / 360 * 2 * np.pi)
            )
            S[:, i] = np.conj(U_0) * U_0
    return S


def buildBusToLineMatrix(grid):
    # Assuming one slack and radial grid

    numberofnodes = len(grid["nodes"])
    numberoflines = len(grid["edges"])

    busToLine = np.zeros((numberoflines, numberofnodes), dtype=np.bool)

    # Prepare an adjacency list of the graph structure
    adj = {node: set() for node in [n["id"] for n in grid["nodes"]]}

    # Step 1: Build Branch-Bus Matrix
    for line in grid["edges"]:
        adj[line["from"]].add((line["id"], line["to"]))
        adj[line["to"]].add((line["id"], line["from"]))

    paths = [[(None, 0)]]
    while paths:
        path = paths.pop()
        # print('Path:',path)
        nodes_in_path = [p[1] for p in path]
        # print('Nodes in Path:',nodes_in_path)
        possible_paths = [p for p in adj[path[-1][1]] if p[1] not in nodes_in_path]
        for next_path in possible_paths:
            paths.append(path + [next_path])
            for line in tuple(p[0] for p in path if p[0] is not None) + (next_path[0],):
                # print('Current from node',next_path[1],'is flowing though line',line)
                busToLine[line, next_path[1]] = True
        # print('--------------------')

    # print(busToLine.astype(np.int16))
    return busToLine[:, 1:].astype(np.complex128)


def buildZall(grid):
    Zall = np.zeros(len(grid["edges"]), dtype=np.complex128)
    for line in grid["edges"]:
        Zall[line["id"]] = line["R"] + 1j * line["X"]
    return Zall


if __name__ == "__main__":
    import data.mockgrids as mockgrids
    from pprint import pprint as pprint

    grid = mockgrids.feeder(20)
    print(grid)
    prepdata(grid)

    # starttime = time.time()
    # buildBusToLineMatrix(grid)
    # print(time.time() - starttime)



import queue
import numpy as np

# import powerflow


def find_impedance_paths(grid):
    """Find impedance from slack to every node in a radial grid
    Parameters:
        grid: [dict] grid data structure containing nodes and edges
    Returns:
        impedance_to_nodes: [dict] <node_name>: [complex] Impedance in Ohms
        path_to_nodes: [dict] <node_name>: [list] of node_names in the path from slack to node
    """

    nodes = grid["nodes"]
    visited = {node["name"]: False for node in nodes}

    impedance_to_nodes = {node["name"]: 0 for node in nodes}
    path_to_nodes = {node["name"]: [] for node in nodes}

    # Use node names, because the ids change during the reduction
    node_id_to_node_name = {n["id"]: n["name"] for n in nodes}

    # Build a dict which contains the neighbours for every node:
    adjacency_dict = {node["name"]: [] for node in nodes}
    for line in grid["edges"]:
        target_name = node_id_to_node_name[line["target"]]
        source_name = node_id_to_node_name[line["source"]]
        adjacency_dict[source_name].append((target_name, line["R"] + 1j * line["X"]))
        adjacency_dict[target_name].append((source_name, line["R"] + 1j * line["X"]))

    # Find Slack node:
    slack_node_names = [node["name"] for node in nodes if node["is_slack"]]
    if len(slack_node_names) == 1:
        slack_node_name = slack_node_names[0]
    else:
        print("0 or 2 or more slack nodes!")
        return None

    # Find adjacent nodes and fill impedances_of_paths using BFS:
    # ===========================================================

    q = queue.Queue()
    q.put((slack_node_name, 0, 0, []))

    while q.qsize():
        node_name, total_path_impedance, line_impedance, node_path = q.get()
        if visited[node_name]:
            continue
        impedance_to_node = total_path_impedance + line_impedance
        impedance_to_nodes[node_name] = impedance_to_node
        path_to_nodes[node_name] = node_path + [node_name]
        # print(f'{node_name}: {round(impedance_to_node.real*1000,1)} mOhm, Path: {" -> ".join([str(i) for i in node_path])}')
        visited[node_name] = True
        for next_node_name, line_impedance in adjacency_dict[node_name]:
            # print(f'\t{next_node_name}: {round(line_impedance.real*1000,1)} mOhm {"-> visited" if visited[next_node_name] else ""}')
            if visited[next_node_name]:
                continue
            q.put(
                (
                    next_node_name,
                    impedance_to_node,
                    line_impedance,
                    node_path + [node_name],
                )
            )

    # print(impedance_to_nodes)

    return impedance_to_nodes, path_to_nodes


def find_candidates_end(grid, S, U0=400 * 0.9, excluded_node_names=[]):
    nodes = grid["nodes"]
    lines = grid["edges"]
    candidates = {}

    for node in nodes:
        node_id = node["id"]

        # Check excluded_nodes
        # for reduction to a single feeder,
        # the nodes in the feeder with the highest impedance path must not be reduced:
        if "name" in node and node["name"] in excluded_node_names:
            continue

        # Shortcut continue for slacks, obviously not removable:
        if node["is_slack"]:
            continue

        # Find connected lines and nodes:
        connected_nodes = []
        connected_lines = []
        for line_i, line in enumerate(lines):
            if node_id == line["source"]:
                connected_nodes.append(line["target"])
                connected_lines.append(line)
            elif node_id == line["target"]:
                connected_nodes.append(line["source"])
                connected_lines.append(line)

        # End Node -> Delete Node and Line:
        if len(connected_nodes) == 1 and connected_nodes[0]:
            candidates[node_id] = {}
            candidates[node_id]["node"] = node
            candidates[node_id]["connected_node_ids"] = connected_nodes
            candidates[node_id]["connected_lines"] = connected_lines
            maxS = max(abs(S[:, node_id]))
            Z = connected_lines[0]["R"] + 1j * connected_lines[0]["X"]
            Uexp = Z * np.conj(maxS / U0)
            # print(f'Node: {node_id}, R={round(Z.real,6)}, X={round(Z.imag,6)}, maxS({node_id})={round(maxS)} -> Uexp={round(abs(Uexp),4)} V')
            candidates[node_id]["Uexp"] = Uexp
    return candidates


def reduce_one_node_end(grid, S, U0=400 * 0.9, excluded_node_names=[]):

    candidates = find_candidates_end(
        grid, S, U0=U0, excluded_node_names=excluded_node_names
    )
    if not candidates:
        return grid, S, {}

    best_candidate = list(sorted(candidates.values(), key=lambda c: abs(c["Uexp"])))[0]
    # print(f'Expected voltage error: {best_candidate["Uexp"]}')

    nodes = grid["nodes"]
    lines = grid["edges"]
    node = best_candidate["node"]

    parent_node_id = best_candidate["connected_node_ids"][0]
    line = best_candidate["connected_lines"][0]

    node_reduced = best_candidate["node"]
    nodes.remove(node_reduced)
    lines.remove(line)

    grid = {"nodes": nodes, "edges": lines, "grid": grid["grid"]}

    S[:, parent_node_id] += (
        S[:, node["id"]]
        + (line["R"] + 1j * line["X"]) * np.conj(abs(S[:, node["id"]]) / abs(U0)) ** 2
    )
    grid, S = powerflow.grid_reduction.lossless.normalise_node_ids(grid, S)
    S = S[:, : len(nodes)]

    return grid, S, node_reduced


def find_candidates_transit(grid, S, excluded_node_names=[]):

    nodes = grid["nodes"]
    lines = grid["edges"]
    candidates = {}

    for node in nodes:
        node_id = node["id"]

        # Check excluded_nodes
        # for reduction to a single feeder,
        # the nodes in the feeder with the highest impedance path must not be reduced:
        if "name" in node and node["name"] in excluded_node_names:
            continue

        # Shortcut continue for slacks, obviously not removable:
        if node["is_slack"]:
            continue

        # Find connected lines and nodes:
        connected_nodes = []
        connected_lines = []
        for line_i, line in enumerate(lines):
            if node_id == line["source"]:
                connected_nodes.append(line["target"])
                connected_lines.append(line)
            elif node_id == line["target"]:
                connected_nodes.append(line["source"])
                connected_lines.append(line)

        if len(connected_nodes) == 2 and connected_nodes[0]:
            candidates[node_id] = {}
            candidates[node_id]["node"] = node
            candidates[node_id]["connected_node_ids"] = connected_nodes
            candidates[node_id]["connected_lines"] = connected_lines
            candidates[node_id]["maxS"] = max(abs(S[:, node_id]))

    return candidates


def reduce_one_node_transit(grid, S, excluded_node_names=[]):
    candidates = find_candidates_transit(
        grid, S, excluded_node_names=excluded_node_names
    )
    if not candidates:
        return grid, S, {}

    nodes = grid["nodes"]
    lines = grid["edges"]

    best_candidate = list(sorted(candidates.values(), key=lambda c: abs(c["maxS"])))[0]

    # Create new line 13:
    connected_lines = best_candidate["connected_lines"]
    new_line = {}
    new_line["id"] = connected_lines[0]["id"]
    new_line["source"] = best_candidate["connected_node_ids"][0]
    new_line["target"] = best_candidate["connected_node_ids"][1]
    new_line["R"] = connected_lines[0]["R"] + connected_lines[1]["R"]
    new_line["X"] = connected_lines[0]["X"] + connected_lines[1]["X"]
    if "length_m" in connected_lines[0] and "length_m" in connected_lines[1]:
        new_line["length_m"] = (
            connected_lines[0]["length_m"] + connected_lines[1]["length_m"]
        )

    # Change S at nodes 1 and 3:
    node1_id = best_candidate["connected_node_ids"][0]
    node2_id = best_candidate["node"]["id"]
    node3_id = best_candidate["connected_node_ids"][1]
    if (
        best_candidate["connected_lines"][0]["source"] == node1_id
        or best_candidate["connected_lines"][0]["target"] == node1_id
    ):

        line12 = best_candidate["connected_lines"][0]
        line23 = best_candidate["connected_lines"][1]
    else:
        line12 = best_candidate["connected_lines"][1]
        line23 = best_candidate["connected_lines"][0]

    Z12 = line12["R"] + 1j * line12["X"]
    Z23 = line23["R"] + 1j * line23["X"]

    S[:, node1_id] += Z23 / (Z12 + Z23) * S[:, node2_id]
    S[:, node3_id] += Z12 / (Z12 + Z23) * S[:, node2_id]

    # Delete Node 2 and lines 12 and 23 and append new line:
    node_reduced = best_candidate["node"]
    nodes.remove(best_candidate["node"])
    for line in connected_lines:
        lines.remove(line)
    lines.append(new_line)
    grid = {"nodes": nodes, "edges": lines, "grid": grid["grid"]}

    grid, S = powerflow.grid_reduction.lossless.normalise_node_ids(grid, S)

    S = S[:, : len(grid["nodes"])]

    return grid, S, node_reduced


def reduce_to_n_nodes(grid, S, n):

    nodes_reduced = []
    while len(grid["nodes"]) > n:

        # First, try intermediate reduction, then end reduction
        grid, S, node_reduced = reduce_one_node_transit(grid, S)
        if node_reduced:
            # print(f'Reducing node {node_reduced["id"]} with transit reduction')
            nodes_reduced.append(node_reduced)
            continue
        else:
            grid, S, node_reduced = reduce_one_node_end(grid, S)
            if not node_reduced:
                break
            # print(f'Reducing node {node_reduced["id"]} with end reduction')
            nodes_reduced.append(node_reduced)

    return grid, S, nodes_reduced


def reduce_to_single_feeder(grid, S):

    impedance_of_paths, path_to_nodes = find_impedance_paths(grid)
    largest_impedance_path = sorted(
        impedance_of_paths.items(), key=lambda i: abs(i[1])
    )[-1]
    final_node_name, impedance_to_slack = largest_impedance_path

    path_of_largest_impedance = path_to_nodes[final_node_name]

    nodes_reduced = []

    # First, try intermediate reduction, then end reduction
    while True:
        grid, S, node_reduced = reduce_one_node_transit(
            grid, S, excluded_node_names=path_of_largest_impedance
        )
        if node_reduced:
            nodes_reduced.append(node_reduced)
            continue
        else:
            grid, S, node_reduced = reduce_one_node_end(
                grid, S, excluded_node_names=path_of_largest_impedance
            )
            if not node_reduced:
                break
            nodes_reduced.append(node_reduced)

    return grid, S, nodes_reduced


def plot_voltage_comparison(U, U_org, filename=None):
    # Plot a comparison of minimal voltage at all 1440 times:
    import matplotlib.ticker as ticker

    minU = np.min(abs(U), axis=1)
    minUorg = np.min(abs(U_org), axis=1)
    fig, (ax1, ax2) = plt.subplots(2, 1)
    plt.setsize(fig, size=1.5)
    ax1.plot(minU)
    ax1.plot(minUorg)
    ax1.set_ylabel("Min. abs. Voltage / V")
    ax1.set_xlabel("Time of day / hh:mm")
    ax1.xaxis.set_major_locator(
        ticker.FixedLocator([t * 60 for t in (0, 6, 12, 18, 24)])
    )
    ax1.xaxis.set_major_formatter(ticker.FuncFormatter(lambda x, pos: f"{x//60}:00"))
    ax1.grid()
    ax2.plot(abs(minU - minUorg))
    ax2.set_ylabel("Difference / V")
    ax2.set_xlabel("Time of day / hh:mm")
    ax2.xaxis.set_major_locator(
        ticker.FixedLocator([t * 60 for t in (0, 6, 12, 18, 24)])
    )
    ax2.xaxis.set_major_formatter(ticker.FuncFormatter(lambda x, pos: f"{x//60}:00"))
    ax2.grid()
    plt.tight_layout()
    if filename:
        plt.save(fig, filename + ".eps")
    plt.show()


def plotgraph_highlight_path(
    grid,
    filename="",
    S=None,
    ax=None,
    verbose=False,
    nodes_in_path=[],
    **additional_args,
):
    import copy
    import networkx as nx
    import powerflow.plotting as plt

    grid = copy.deepcopy(grid)
    g = nx.Graph()
    nodes = grid["nodes"]
    node_ids = [g["id"] for g in nodes]
    if verbose:
        print(f"Plotting {len(node_ids)} Nodes...")
    positions = {n["id"]: (n["x"], n["y"]) for n in nodes}
    g.add_nodes_from(node_ids)
    for e in grid["edges"]:
        if e["source"] in nodes_in_path and e["target"] in nodes_in_path:
            color = "darkslategray"
        else:
            color = "lightgray"
        g.add_edge(e["source"], e["target"], color=color)

    if S is not None:
        for i, n in enumerate(nodes):
            if np.sum(S[:, n["id"]]) != 0:
                nodes[i]["noload"] = False
            else:
                nodes[i]["noload"] = True

    # Set Slacks to green, loads to orange, rest to gray:
    node_color = [
        "darkslategray" if n["id"] in nodes_in_path else "lightgray"
        for n in grid["nodes"]
    ]
    node_color = [
        "orangered" if not n["noload"] else nc
        for n, nc in zip(grid["nodes"], node_color)
    ]
    node_color = [
        "green" if n["is_slack"] else nc for n, nc in zip(grid["nodes"], node_color)
    ]

    node_size = [18 if not n["noload"] else 3 for n in grid["nodes"]]
    node_size = [18 if n["is_slack"] else ns for n, ns in zip(grid["nodes"], node_size)]

    edge_colors = [g[i][j]["color"] for i, j in g.edges]

    if ax == None:
        fig, ax = plt.subplots()
        plt.setsize(fig, 1)
        ax.axis("off")
        plt.margins(0.01)
        ax.margins(0.02)
        showplot = True
    else:
        showplot = False

    nx.draw_networkx(
        g,
        ax=ax,
        pos=positions,
        node_color=node_color,
        node_size=node_size,
        edge_color=edge_colors,
        width=1.5,
        with_labels=False,
        **additional_args,
    )

    plt.tight_layout()

    if showplot:
        if filename:
            plt.save(fig, f"{filename}.eps")
        plt.show()


if __name__ == "__main__":

    import sys
    import json
    import powerflow
    import powerflow.plotting as plt

    # Load original grid model:
    grid = powerflow.grids.eu_lv_feeder()
    S = powerflow.loads.eu_lv_feeder(grid)
    # plt.plotgraph(grid)

    # Lossless Reduction:
    grid, reduced_nodes, S = powerflow.grid_reduction.lossless.reduce(grid, S)
    # U_org, iters, runtime = powerflow.zbusjacobi(grid, S, verbose=True)
    # plt.plotgraph(grid)

    # Reduction to n nodes:
    n = 5
    print(f"\nReducing grid to {n} nodes")
    grid_red, S_red, nodes_reduced = reduce_to_n_nodes(grid, S, n)
    # U_red, iters_red, runtime_red = powerflow.zbusjacobi(grid_red, S_red)
    # U_red, iters_red, runtime_red = powerflow.ybusnewton(grid_red, S_red)
    # U_red, iters_red, runtime_red = powerflow.ybusgaussseidel(grid_red, S_red)
    # U_red, iters_red, runtime_red = powerflow.ybusjacobi(grid_red, S_red)
    plt.plotgraph(grid_red, S=S_red, filename=f"european_lv_feeder_red_to_{n}_2")
    # plot_voltage_comparison(U_red, U_org, f'european_lv_feeder_voltages_red_to_{n}')

    # Reduction to single feeder:
    # print(f'\nReducing grid to single feeder')
    # grid_feeder, S_feeder, nodes_reduced = reduce_to_single_feeder(grid, S)
    # U_feeder, iters_feeder, runtime_feeder = powerflow.ybusnewton(grid_feeder, S_feeder)
    # U_feeder, iters_feeder, runtime_feeder = powerflow.zbusjacobi(grid_feeder, S_feeder)
    # U_feeder, iters_feeder, runtime_feeder = powerflow.ybusgaussseidel(grid_feeder, S_feeder)
    # U_feeder, iters_feeder, runtime_feeder = powerflow.ybusjacobi(grid_feeder, S_feeder)
    # U_feeder, iters_feeder, runtime_feeder = powerflow.bfs(grid_feeder, S_feeder)
    # plt.plotgraph(grid_feeder,S=S_feeder)
    # plot_voltage_comparison(U_feeder, U_org)
    # plt.plotgraph(grid_feeder,S=S_feeder, filename='european_lv_feeder_singlefeeder')
    # plot_voltage_comparison(U_feeder, U_org, 'european_lv_feeder_voltages_singlefeeder')

    # Plot grid with path of highest impedance highlighted:
    # impedance_of_paths, path_to_nodes = find_impedance_paths(grid)
    # largest_impedance_path = sorted(impedance_of_paths.items(), key=lambda i:abs(i[1]))[-1]
    # final_node_name, _ = largest_impedance_path
    # path_of_largest_impedance = path_to_nodes[final_node_name]
    # node_id_to_node_name = {n['name']:n['id'] for n in grid['nodes']}
    # nodes_in_path = [node_id_to_node_name[node_name] for node_name in path_of_largest_impedance]

    # plotgraph_highlight_path(grid, S=S, nodes_in_path=nodes_in_path, filename='european_lv_feeder_highlightpath')



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['fastpf', 'fastpf.data', 'fastpf.grid_reduction', 'fastpf.plotting']

package_data = \
{'': ['*'],
 'fastpf': ['.pytest_cache/*', '.pytest_cache/v/*', '.pytest_cache/v/cache/*']}

install_requires = \
['loguru >= 0.6.0',
 'numpy >= 1.16.0',
 'scipy >= 1.8.0',
 'numba >= 0.55.1',
 'pathos >= 0.2.8',
 'threadpoolctl >= 3.1.0']

setup(name='fastpf',
      version='0.0.5',
      description='Performance-first Power Flow implemented in Python with AOT Compilation',
      author=None,
      author_email='Christoph Kattmann <christoph.kattmann@gmail.com>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
     )



import json
import queue
import copy
import numpy as np
import networkx as nx

import pprint

def reduce(grid, S=None, verbose=False):
    ''' Remove electrically irrelevant nodes from a grid
    '''

    if verbose:
        log = print
    else:
        log = lambda *args, **kwargs: None

    # Avoid changing the grid in-place:
    grid = copy.deepcopy(grid)

    nodes = grid['nodes']
    lines = grid['edges']
    node_deleted_last_run = True
    nodes_deleted = []
    # Try deleting nodes until no unnecessary node left:
    while node_deleted_last_run:
        node_deleted_last_run = False
        for node_i, node in enumerate(nodes):
            node_id = node['id']

            # Shortcut continue for loads and slacks, obviously not removable:
            if 'is_slack' in node.keys() and node['is_slack']:
                log(f'Node {node["id"]} is slack, skipping')
                continue
            if S is None:
                if 'noload' in node.keys() and node['noload'] == False:
                    continue
            else:
                if np.any(np.abs(S[:,node['id']]) > 0):
                    log(f'Node {node["id"]} has load {np.max(np.abs(S)[:,node["id"]])}, skipping')
                    continue

            # Find connected lines and nodes:
            connected_nodes = []
            connected_lines = []
            line_indices = []
            for line_i, line in enumerate(lines):
                if node['id'] == line['source']:
                    connected_nodes.append(line['target'])
                    connected_lines.append(line)
                    line_indices.append(line_i)
                elif node['id'] == line['target']:
                    connected_nodes.append(line['source'])
                    connected_lines.append(line)
                    line_indices.append(line_i)

            # End Node -> Delete Node and Line:
            if len(connected_nodes) == 1:
                log(f'Deleting End Node {node["id"]}')
                nodes_deleted.append(node['id'])
                del nodes[node_i]
                del lines[line_indices[0]]
                node_deleted_last_run = True

            # Transit Node -> Add up connected lines and delete node:
            elif len(connected_nodes) == 2:
                assert len(connected_lines) == 2
                log(f'Deleting Transit Node {node["id"]}')
                nodes_deleted.append(node['id'])
                new_line = {}
                new_line['id'] = connected_lines[0]['id']
                new_line['source'] = connected_nodes[0]
                new_line['target'] = connected_nodes[1]
                if 'length_m' in connected_lines[0].keys() and 'length_m' in connected_lines[1].keys():
                    new_line['length_m'] = connected_lines[0]['length_m'] + connected_lines[1]['length_m']
                new_line['R'] = connected_lines[0]['R'] + connected_lines[1]['R']
                new_line['X'] = connected_lines[0]['X'] + connected_lines[1]['X']

                del nodes[node_i]
                for i in reversed(sorted(line_indices)):
                    del lines[i]
                lines.append(new_line)
                node_deleted_last_run = True

            else:
                log(f'Node {node["id"]} has {len(connected_nodes)} connected nodes, skipping')


    # Construct new grid dict:
    reduced_grid = {'nodes': nodes, 'edges': lines, 'grid': grid['grid']}

    # return reduced_grid, 0, S
    reduced_grid_normalised, S_normalized = normalise_node_ids(reduced_grid, S)

    # Delete reduced nodes from loads S:
    if S is not None:
        S_normalized_reduced = S_normalized[:,:len(nodes)]
    else:
        S_normalized_reduced = None
    return reduced_grid_normalised, nodes_deleted, S_normalized_reduced




def normalise_node_ids(grid, S=None):

    grid = copy.deepcopy(grid)

    nodes = grid['nodes']
    lines = grid['edges']

    node_id_dict = {n['id']:n for n in nodes}
    old_id_to_new_id = {}

    node_queue = queue.Queue()
    handled_nodes = []
    node_ids_in_queue = []
    node_id_counter = 0

    # Find first slack node to start:
    # print(nodes)
    for n in nodes:
        if n['is_slack']:
            node_queue.put(n)
            break

    while node_queue.qsize():
        node = node_queue.get()
        node_id = node['id']
        if node_id in handled_nodes:
            continue
        handled_nodes.append(node_id)
        for line in lines:
            if line['source'] == node_id:
                node_queue.put(node_id_dict[line['target']])
                node_ids_in_queue.append(line['target'])
            if line['target'] == node_id:
                node_queue.put(node_id_dict[line['source']])
                node_ids_in_queue.append(line['target'])
        old_id_to_new_id[node_id] = node_id_counter
        node_id_counter += 1

    if S is not None:
        Snew = np.zeros_like(S)
        for n in nodes:
            Snew[:,old_id_to_new_id[n['id']]] = S[:,n['id']]
    else:
        Snew = None

    for n in nodes:
        n['id'] = old_id_to_new_id[n['id']]

    for line_id, line in enumerate(lines):
        line['id'] = line_id
        line['source'] = old_id_to_new_id[line['source']]
        line['target'] = old_id_to_new_id[line['target']]

    grid = {'nodes': nodes, 'edges': lines, 'grid': grid['grid']}

    return grid, Snew






### DEV CODE ###

def reduce_intermediate_nodes(grid):
    nodes = grid['nodes']
    lines = grid['edges']
    node_deleted = True
    nodes_reduced = []
    while node_deleted:
        node_deleted = False
        for node_i, node in enumerate(nodes):  # reversed because generally, "chains" of reducable nodes start at the end
            if not node['noload'] or node['is_slack']:
                continue
            connected_nodes = []
            connected_lines = []
            line_indices = []
            for line_i, line in enumerate(lines):
                if node['id'] == line['source']:
                    connected_nodes.append(line['target'])
                    connected_lines.append(line)
                    line_indices.append(line_i)
                elif node['id'] == line['target']:
                    connected_nodes.append(line['source'])
                    connected_lines.append(line)
                    line_indices.append(line_i)

            if len(connected_nodes) == 2 and node['noload'] and not node['is_slack']:
                nodes_reduced.append(node['id'])
                new_line = {}
                new_line['id'] = connected_lines[0]['id']
                new_line['source'] = connected_nodes[0]
                new_line['target'] = connected_nodes[1]
                new_line['length_m'] = connected_lines[0]['length_m'] + connected_lines[1]['length_m']
                new_line['R'] = connected_lines[0]['R'] + connected_lines[1]['R']
                new_line['X'] = connected_lines[0]['X'] + connected_lines[1]['X']

                # Actually delete the nodes:
                del nodes[node_i]
                for i in reversed(sorted(line_indices)):
                    del lines[i]
                lines.append(new_line)
                node_deleted = True
                break
    grid = {'nodes': nodes, 'edges': lines, 'grid': grid['grid']}

    return grid, nodes_reduced



def reduce_end_nodes(grid):
    nodes = grid['nodes']
    lines = grid['edges']
    node_deleted = True
    nodes_reduced = []
    while node_deleted:
        node_deleted = False
        for node_i, node in enumerate(nodes):  # reversed because generally,
            connected_nodes = []
            line_indices = []
            for line_i, line in enumerate(lines):
                if node['id'] == line['source']:
                    connected_nodes.append(line['target'])
                    line_indices.append(line_i)
                elif node['id'] == line['target']:
                    connected_nodes.append(line['source'])
                    line_indices.append(line_i)
                if len(connected_nodes) > 1:
                    break

            if len(connected_nodes) == 1 and node['noload'] and not node['is_slack']:
                nodes_reduced.append(node['id'])
                del nodes[node_i]
                del lines[line_indices[0]]
                node_deleted = True
                break

    grid = {'nodes': nodes, 'edges': lines, 'grid': grid['grid']}
    return grid, nodes_reduced



def reduce_grid(grid):

    nodes_reduced_end = 1
    nodes_reduced_intermediate = 1

    while nodes_reduced_intermediate  or nodes_reduced_end:
        print('End Reduction...')
        grid, nodes_reduced_end = reduce_end_nodes(grid)
        print(f'Reduced {nodes_reduced_end} Nodes by end_reduction -> Nodes: {len(grid["nodes"])}, Lines: {len(grid["edges"])}')
        if nodes_reduced_end:
            plotgraph(grid)

        print('Intermediate Reduction...')
        grid, nodes_reduced_intermediate = reduce_intermediate_nodes(grid)
        print(f'Reduced {nodes_reduced_intermediate} Nodes by intermediate_reduction -> Nodes: {len(grid["nodes"])}, Lines: {len(grid["edges"])}')
        if nodes_reduced_intermediate:
            plotgraph(grid)

    return grid

if __name__ == '__main__':

    with open('european_lv_feeder.json','r') as f:
        grid = json.load(f)

    print(f'Nodes: {len(grid["nodes"])}, Lines: {len(grid["edges"])}')

    plotgraph(grid, filename='europeanLvFeeder_nonreduced')

    grid, nodes_reduced = reduce_grid(grid)

    plotgraph(grid, filename='europeanLvFeeder_reduced')

    with open('european_lv_feeder_reduced.json', 'w') as f:
        json.dump(grid, f)




"""
Performance-first Power Flow implemented in Python with AOT Compilation
"""

__version__ = "0.0.5"

from .log import logger
from .log import set_loglevel

from . import powerflow_methods

try:
    from . import powerflow_methods_cc
except ImportError:
    powerflow_methods.compile()
    from . import powerflow_methods_cc

from .data.parse_matpower_casefile import parse_matpower_casefile
from .data import testgrids
from .data import testloads

from .validation import validate_grid
from .process_grid import process_grid
from .process_grid import integrate_slacks_for_Yident

from .powerflow import ybusjacobi
from .powerflow import ybusgaussseidel
from .powerflow import ybusnewton
from .powerflow import zbusjacobi
from .powerflow import compare_methods

from .plotting import *


set_loglevel("INFO")



import sys
import collections
import numpy as np
from .log import logger


def validate_grid(grid, S=None, logging_enabled=True, raise_on_error=False):
    """Check the grid dict for internal consistency"""

    logger.debug("Checking grid for validity...")
    error_messages = []  # we will now see about that

    def _add_to_errors(message):
        if logging_enabled:
            logger.error(message)
        error_messages.append(message)
        if raise_on_error:
            raise ValueError(message)

    # Check if the grid has a 'nodes'-list:
    if not "nodes" in grid:
        _add_to_errors('No "nodes" in grid')
    nodes = grid["nodes"]

    # Check if the grid has a 'nodes'-list:
    if not "lines" in grid:
        _add_to_errors('No "lines" in grid')
    lines = grid["lines"]

    # Check if the nodes-entries have at least an id:
    for key in ["id"]:
        for node in grid["nodes"]:
            if key not in node:
                _add_to_errors(f'{key} missing in grid["nodes"] for node {node}')

    # Check if the lines-entries have at least the necessary keys:
    for line in lines:
        for key in ["id", "from", "to", "R", "X"]:
            if key not in line:
                _add_to_errors(f'"{key}" missing in grid["lines"] for line {line}')

    # Check if node indices are properly sorted:
    number_of_nodes = len(nodes)
    node_index_list = []
    for node in nodes:
        node_index_list.append(node["id"])
    for n in range(number_of_nodes):
        if n not in node_index_list:
            _add_to_errors(f"Node ID {n} not found")
    for i in node_index_list:
        if n not in range(number_of_nodes):
            _add_to_errors(f"Node ID {n} not found")

    # Check if there are slack nodes in the grid:
    slacknodes = [n for n in nodes if n["is_slack"]]
    if not slacknodes:
        _add_to_errors("No slack node found in grid")

    # Check if slack nodes have a slack_voltage and slack_angle:
    for sn in slacknodes:
        if "slack_voltage" not in sn:
            _add_to_errors(
                'is_slack is True for Node {sn["id"]}, but no slack voltage is given'
            )
        if "slack_angle" not in sn:
            _add_to_errors(
                f'is_slack is True for Node {sn["id"]}, but no slack angle is given'
            )

    # Check if any of the lines lead to a node that does not exist:
    node_ids = [n["id"] for n in nodes]
    for line in lines:
        if line["from"] not in node_ids:
            _add_to_errors(
                f'Line {line["id"]} has source {line["from"]}, which is not a valid Node ID}}'
            )
        if line["to"] not in node_ids:
            _add_to_errors(
                f'Line {line["id"]} has target {line["to"]}, which is not a valid Node ID}}'
            )

    # Check if every node has a line attached to it:
    for n in nodes:
        connected = False
        for line in lines:
            if line["from"] == n["id"]:
                connected = True
                break
            if line["to"] == n["id"]:
                connected = True
                break
        if not connected:
            _add_to_errors(f"Node {n['id']} is not connected")

    # Check if S works with grid:
    if S is not None:
        if type(S) is not np.ndarray:
            _add_to_errors("type(S) is not numpy.ndarray")

        numberofloads, numberofnodes = S.shape
        if numberofnodes != len(nodes):
            _add_to_errors(
                f"Number of nodes between grid(n={len(nodes)}) and S(n={numberofnodes}) inconsistent"
            )

    if not error_messages:
        logger.debug("Validation: All checks passed")

    return error_messages


def is_radial(grid):

    # Shortcut:
    if len(grid["nodes"]) != len(grid["lines"]) + 1:
        return False

    sourcenodes = [min(l["source"], l["target"]) for l in grid["lines"]]
    sourcenodes_count = collections.Counter(sourcenodes).values()
    if not all([c == 1 for c in sourcenodes_count]):
        return True
    else:
        return False


def determine_gridshape(grid):
    """Determine the topology of the grid - feeder, radial, or meshed"""

    if len(grid["nodes"]) != len(grid["lines"]) + 1:
        is_radial = False
        is_feeder = False
        return is_feeder, is_radial

    # Check if Feeder: If a source node appears more than once, there is a junction
    sourcenodes = [min(l["source"], l["target"]) for l in grid["lines"]]
    sourcenodes_count = collections.Counter(sourcenodes).values()
    if not all([c == 1 for c in sourcenodes_count]):
        is_feeder = False
        is_radial = True
        return is_feeder, is_radial

    is_feeder = True
    is_radial = True

    return is_feeder, is_radial



import warnings
import numpy as np
import scipy.sparse
import numba
from numba.pycc import CC
from numba.core.errors import NumbaPerformanceWarning

cc = CC("powerflow_methods_cc")

warnings.simplefilter("ignore", category=NumbaPerformanceWarning)

np.seterr("raise")

# YBus Jacobi
# ===========
@cc.export(
    "ybusjacobi", "(complex128[:,:],complex128[:], complex128[:,:], float64, int64)"
)
def ybusjacobi(Y, U, S, eps_s=1.0, max_iters=10000):
    Y = Y.copy()
    S = S.copy()
    diagY = np.diag(Y)
    u_all = np.zeros((S.shape[0], U.shape[0]), dtype=np.complex128)
    iters_all = np.zeros(S.shape[0], dtype=np.int32)
    for i in range(S.shape[0]):
        iters = 0
        s = S[i, :]
        while True:
            IR = Y @ U - np.conj(s / U)
            if np.max(np.abs(U * np.conj(IR))) < eps_s or iters > max_iters:
                break
            U -= IR / diagY
            iters += 1
        u_all[i, :] = U
        iters_all[i] = iters
    return u_all, iters_all


# YBus Gauss-Seidel
# =================
@cc.export(
    "ybusgaussseidel",
    "(complex128[:,:], complex128[:], complex128[:,:], float64, float64, int64)",
)
def ybusgaussseidel(Y, U, S, acc_factor=1.0, eps_s=1, max_iters=5000):
    u_all = np.zeros((S.shape[0], U.shape[0]), dtype=np.complex128)
    iters_all = np.zeros(S.shape[0])
    diagY = np.diag(Y)
    for i in range(S.shape[0]):
        s = S[i, :]
        iters = 0
        while True:
            for j in range(Y.shape[0]):
                U[j] -= acc_factor * ((Y[j, :] @ U) - np.conj(s[j] / U[j])) / diagY[j]
            if np.max(np.abs(U * np.conj(Y @ U) - s)) < eps_s or iters > max_iters:
                break
            iters += 1
        u_all[i, :] = U
        iters_all[i] = iters
    return u_all, iters_all


@cc.export(
    "ybusgs_nminus1",
    "(complex128[:,:], int32, int32, complex128, complex128[:], complex128[:], float64, float64, int64)",
)
def gs_nminus1(Y, source_id, target_id, Y_line, s, U, acc_factor, eps_s, max_iters):
    # Switch off one line:
    Y[source_id, target_id] -= Y_line
    Y[target_id, source_id] -= Y_line
    Y[source_id, source_id] += Y_line
    Y[target_id, target_id] += Y_line

    iters = 0
    diagY = np.diag(Y)
    while True:
        for j in range(Y.shape[0]):
            U[j] -= 1.2 * ((Y[j, :] @ U) - np.conj(s[j] / U[j])) / diagY[j]
        if np.max(np.abs(U * np.conj(Y @ U) - s)) < eps_s or iters > max_iters:
            break
        iters += 1
    return U, iters


# YBus Relaxation
# ===============
@cc.export(
    "ybusrelaxation", "(complex128[:,:],complex128[:], complex128[:,:], float64, int64)"
)
def ybusrelaxation(Y, U, S_all, s_eps=0.1, max_iters=20000):
    diagY = np.diag(Y)
    U_all = np.zeros((S_all.shape[0], U.shape[0]), dtype=np.complex128)
    iters_all = np.zeros(S_all.shape[0])
    for i in range(S_all.shape[0]):
        s = S_all[i, :]
        iters = 0
        while True:
            IR = Y @ U - np.conj(s / U)
            SR = np.conj(np.conj(U) * IR)
            max_R_index = np.argmax(np.abs(SR))
            if np.abs(SR[max_R_index]) < s_eps or iters > max_iters:
                break
            iters += 1
            U[max_R_index] -= IR[max_R_index] / diagY[max_R_index]
        U_all[i, :] = U
        iters_all[i] = iters
    return U_all, iters_all


# YBus Newton-Raphson
# ===================
# @cc.export('ybusnewton_old','(complex128[:,:], complex128[:,:], complex128[:], complex128[:,:], int32, float64, int64)')
# def ybusnewton(Y, Yred, U, S_all, slack_index=0, eps_s=1, max_iters=100):

#     numberofnodes = U.shape[0] - 1

#     # This works, but looks stupid:
#     non_slack_indices = np.zeros(numberofnodes, dtype=np.int32)
#     counter = 0
#     for i in range(U.shape[0]):
#         if i != slack_index:
#             non_slack_indices[counter] = i
#             counter += 1

#     S_all = S_all[:,non_slack_indices]

#     U_abs = np.abs(U)[non_slack_indices]
#     # This has to be changed into the given angles:
#     phi = np.zeros(U_abs.size, dtype=np.float64)

#     U_all = np.zeros((S_all.shape[0], U.shape[0]), dtype=np.complex128)
#     iters_all = np.zeros(S_all.shape[0])

#     dPQ = np.zeros(2*S_all.shape[1], dtype=np.float64)
#     J = np.zeros((numberofnodes*2, numberofnodes*2), dtype=np.float64)

#     Yredconj = np.conj(Yred)

#     iters = 0

#     for j in range(S_all.shape[0]):
#         iters = 0
#         s = S_all[j,:]

#         while iters <= max_iters:
#             U[non_slack_indices] = U_abs * np.exp(1j*phi)
#             Scalc = U * (np.conj(Y @ U))
#             Smismatch = Scalc[non_slack_indices] - s
#             Pcalc = Smismatch.real
#             Qcalc = Smismatch.imag
#             dPQ = np.hstack((Pcalc, Qcalc))
#             if np.max(np.abs(dPQ)) < eps_s:
#                 break

#             diagIconj = np.diag(np.conj((Y @ U)[non_slack_indices]))
#             diagUred = np.diag(U[non_slack_indices])
#             Jphi = 1j * diagUred @ (diagIconj - np.conj(diagUred) @ Yredconj)
#             JU = diagIconj + diagUred @ Yredconj

#             J[:numberofnodes, :numberofnodes] = np.real(Jphi)
#             J[numberofnodes:, :numberofnodes] = np.imag(Jphi)
#             J[:numberofnodes, numberofnodes:] = np.real(JU)
#             J[numberofnodes:, numberofnodes:] = np.imag(JU)

#             # dx = np.linalg.inv(-J) @ dPQ
#             dx = np.linalg.solve(-J,dPQ)
#             phi += dx[0:int(dx.shape[0]/2)]
#             U_abs += (dx[int(dx.shape[0]/2):])

#             iters += 1

#         U_all[j] = U
#         iters_all[j] = iters

#     return U_all, iters_all


@cc.export(
    "ybusnewton", "(complex128[:,:], complex128[:,:], complex128[:], float64, int64)"
)
def ybusnewton(Y, S_all, U, eps_s=1, max_iters=100):
    """Assume that node 0 is slack"""

    numberofnodes = U.shape[0] - 1

    S_all = S_all[:, 1:]

    U_abs = np.abs(U)[1:]
    # This has to be changed into the given angles:
    phi = np.zeros(U_abs.size, dtype=np.float64)

    U_all = np.zeros((S_all.shape[0], U.shape[0]), dtype=np.complex128)
    iters_all = np.zeros(S_all.shape[0])

    dPQ = np.zeros(2 * S_all.shape[1], dtype=np.float64)
    J = np.zeros((numberofnodes * 2, numberofnodes * 2), dtype=np.float64)

    Yredconj = np.conj(Y[1:, 1:])

    iters = 0

    for j in range(S_all.shape[0]):
        iters = 0
        s = S_all[j, :]

        while iters <= max_iters:
            U[1:] = U_abs * np.exp(1j * phi)
            Scalc = U * (np.conj(Y @ U))
            Smismatch = Scalc[1:] - s
            Pcalc = Smismatch.real
            Qcalc = Smismatch.imag
            dPQ = np.hstack((Pcalc, Qcalc))
            if np.max(np.abs(dPQ)) < eps_s:
                break

            diagIconj = np.diag(np.conj((Y @ U)[1:]))
            diagUred = np.diag(U[1:])
            Jphi = 1j * diagUred @ (diagIconj - np.conj(diagUred) @ Yredconj)
            JU = diagIconj + diagUred @ Yredconj

            J[:numberofnodes, :numberofnodes] = np.real(Jphi)
            J[numberofnodes:, :numberofnodes] = np.imag(Jphi)
            J[:numberofnodes, numberofnodes:] = np.real(JU)
            J[numberofnodes:, numberofnodes:] = np.imag(JU)

            dx = np.linalg.solve(-J, dPQ)
            phi += dx[0 : int(dx.shape[0] / 2)]
            U_abs += dx[int(dx.shape[0] / 2) :]

            iters += 1

        U_all[j] = U
        iters_all[j] = iters

    return U_all, iters_all


@cc.export(
    "ybusnewton_mslacks",
    "(complex128[:,:], complex128[:,:], complex128[:], int64, float64, int64)",
)
def ybusnewton_mslacks(Y, S_all, U, m, eps_s=1, max_iters=100):
    """YBus-Newton Power Flow for multiple slack nodes
    Y       [n x n]     complex128: Admittance matrix, main_slack in first line, other slacks in subsequent lines
    S_all   [l x n]     complex128: Loads in Watt
    U       [n]         complex128: Starting voltages, slack voltages need to be in place
    m                   int64:      Number of additional slacks
    eps_s               float64:    power residual
    max_iters           int64:      number of iterations before break
    """

    numberofnodes = U.shape[0] - 1

    S_all = S_all[:, 1:]

    U_abs = np.abs(U)[1:]
    phi = np.zeros(U_abs.size, dtype=np.float64)

    U_all = np.zeros((S_all.shape[0], U.shape[0]), dtype=np.complex128)
    iters_all = np.zeros(S_all.shape[0])
    additional_slack_voltages = U[1 : m + 1]
    dPQ = np.zeros(2 * S_all.shape[1], dtype=np.float64)
    J = np.zeros((numberofnodes * 2, numberofnodes * 2), dtype=np.float64)

    Yredconj = np.conj(Y[1:, 1:])

    iters = 0

    for j in range(S_all.shape[0]):
        iters = 0
        s = S_all[j, :]

        while iters <= max_iters:
            U[1:] = U_abs * np.exp(1j * phi)
            Scalc = U * (np.conj(Y @ U))
            Smismatch = Scalc[1:] - s
            dPQ = np.hstack((Smismatch.real, Smismatch.imag))
            for i in range(len(additional_slack_voltages)):
                dPQ[i] = 0
                dPQ[i + numberofnodes] = 0
            if np.max(np.abs(dPQ)) < eps_s:
                break

            diagIconj = np.diag(np.conj((Y @ U)[1:]))
            diagUred = np.diag(U[1:])
            Jphi = 1j * diagUred @ (diagIconj - np.conj(diagUred) @ Yredconj)
            JU = diagIconj + diagUred @ Yredconj

            J[:numberofnodes, :numberofnodes] = np.real(Jphi)
            J[numberofnodes:, :numberofnodes] = np.imag(Jphi)
            J[:numberofnodes, numberofnodes:] = np.real(JU)
            J[numberofnodes:, numberofnodes:] = np.imag(JU)

            # Handle multiple slacks:
            for i in range(len(additional_slack_voltages)):
                J[i, :] = np.zeros(numberofnodes * 2)
                J[i, i] = 1
                i_phase = i + numberofnodes
                J[i_phase, :] = np.zeros(numberofnodes * 2)
                J[i_phase, i_phase] = 1

            # Handle generator nodes:
            # for k in range(len(additional_slack_voltages)):
            #     J[i, :] = np.zeros(numberofnodes * 2)
            #     J[i, i] = 1
            #     i_phase = i + numberofnodes
            #     J[i_phase, :] = np.zeros(numberofnodes * 2)
            #     J[i_phase, i_phase] = 1

            dx = np.linalg.solve(-J, dPQ)
            phi += dx[0 : int(dx.shape[0] / 2)]
            U_abs += dx[int(dx.shape[0] / 2) :]

            iters += 1

        U_all[j] = U
        iters_all[j] = iters

    return U_all, iters_all


@cc.export(
    "ybusnewton_full",
    "(complex128[:,:], complex128[:,:], complex128[:], int64, int64, float64, int64)",
)
def ybusnewton_full(Y, S_all, U, n_slack, n_pv, eps_s=1, max_iters=100):
    """YBus-Newton Power Flow for multiple slack nodes
    Y       [n x n]     complex128: Admittance matrix, main_slack in first line, other slacks in subsequent lines
    S_all   [l x n]     complex128: Loads in Watt
    U       [n]         complex128: Starting voltages, slack voltages need to be in place
    n_slack             int64:      Number of additional slacks
    n_pv                int32:      Number of PV nodes, sorted to the end of Y
    # U_pv  [n_pv]      float64:    Fixed voltage magnitude setpoints for generators
    eps_s               float64:    power residual
    max_iters           int64:      number of iterations before break
    """

    numberofnodes = U.shape[0] - 1

    S_all = S_all[:, 1:]

    U_abs = np.abs(U)[1:]
    phi = np.zeros(U_abs.size, dtype=np.float64)

    U_all = np.zeros((S_all.shape[0], U.shape[0]), dtype=np.complex128)
    iters_all = np.zeros(S_all.shape[0])
    additional_slack_voltages = U[1 : n_slack + 1]
    dPQ = np.zeros(2 * S_all.shape[1], dtype=np.float64)
    J = np.zeros((numberofnodes * 2, numberofnodes * 2), dtype=np.float64)

    Yredconj = np.conj(Y[1:, 1:])

    iters = 0

    for j in range(S_all.shape[0]):
        iters = 0
        s = S_all[j, :]

        while iters <= max_iters:
            U[1:] = U_abs * np.exp(1j * phi)
            Scalc = U * (np.conj(Y @ U))
            Smismatch = Scalc[1:] - s
            dPQ = np.hstack((Smismatch.real, Smismatch.imag))
            for i in range(len(additional_slack_voltages)):
                dPQ[i] = 0
                dPQ[i + numberofnodes] = 0
            if np.max(np.abs(dPQ)) < eps_s:
                break

            diagIconj = np.diag(np.conj((Y @ U)[1:]))
            diagUred = np.diag(U[1:])
            Jphi = 1j * diagUred @ (diagIconj - np.conj(diagUred) @ Yredconj)
            JU = diagIconj + diagUred @ Yredconj

            J[:numberofnodes, :numberofnodes] = np.real(Jphi)
            J[numberofnodes:, :numberofnodes] = np.imag(Jphi)
            J[:numberofnodes, numberofnodes:] = np.real(JU)
            J[numberofnodes:, numberofnodes:] = np.imag(JU)

            # Handle multiple slacks:
            for i in range(len(additional_slack_voltages)):
                J[i, :] = np.zeros(numberofnodes * 2)
                J[i, i] = 1
                i_phase = i + numberofnodes
                J[i_phase, :] = np.zeros(numberofnodes * 2)
                J[i_phase, i_phase] = 1

            # Handle generator nodes:
            # for k in range(len(additional_slack_voltages)):
            #     J[i, :] = np.zeros(numberofnodes * 2)
            #     J[i, i] = 1
            #     i_phase = i + numberofnodes
            #     J[i_phase, :] = np.zeros(numberofnodes * 2)
            #     J[i_phase, i_phase] = 1

            dx = np.linalg.solve(-J, dPQ)
            phi += dx[0 : int(dx.shape[0] / 2)]
            U_abs += dx[int(dx.shape[0] / 2) :]

            iters += 1

        U_all[j] = U
        iters_all[j] = iters

    return U_all, iters_all


def ybusnewton_sparse(Y, Yred, U, S, slack_index=0, eps_s=1, max_iters=100):

    numberofnodes = U.shape[0] - 1

    # This looks stupid, but works and is outside the hotpath:
    non_slack_indices = np.zeros(numberofnodes, dtype=np.int32)
    counter = 0
    for i in range(S.shape[1]):
        if i != slack_index:
            non_slack_indices[counter] = i
            counter += 1

    S = S[:, non_slack_indices]

    U_abs = np.abs(U)[non_slack_indices]
    # This has to be changed into the given angles:
    phi = np.zeros(U_abs.size, dtype=np.float64)

    u_all = np.zeros((S.shape[0], U.shape[0]), dtype=np.complex128)
    iters_all = np.zeros(S.shape[0])

    dPQ = np.zeros(2 * S.shape[1], dtype=np.float64)
    J = np.zeros((numberofnodes * 2, numberofnodes * 2), dtype=np.float64)

    Yredconj = np.conj(Yred)

    iters = 0

    # Convert to CSR sparse:
    Y = scipy.sparse.csr_matrix(Y)

    for j in range(S.shape[0]):
        iters = 0
        s = S[j, :]

        while iters <= max_iters:
            U[non_slack_indices] = U_abs * np.exp(1j * phi)
            Scalc = U * (np.conj(Y @ U))
            Smismatch = Scalc[non_slack_indices] - s
            Pcalc = Smismatch.real
            Qcalc = Smismatch.imag
            dPQ = np.hstack((Pcalc, Qcalc))
            if np.max(np.abs(dPQ)) < eps_s:
                break
            iters += 1

            diagIconj = np.diag(np.conj((Y.dot(U))[non_slack_indices]))
            # diagIconj = np.diag(np.conj((Y @ U)[non_slack_indices]))
            diagUred = np.diag(U[non_slack_indices])
            Jphi = 1j * diagUred @ (diagIconj - np.conj(diagUred) @ Yredconj)
            JU = diagIconj + diagUred @ Yredconj

            J[:numberofnodes, :numberofnodes] = np.real(Jphi)
            J[numberofnodes:, :numberofnodes] = np.imag(Jphi)
            J[:numberofnodes, numberofnodes:] = np.real(JU)
            J[numberofnodes:, numberofnodes:] = np.imag(JU)

            J = scipy.sparse.csr_matrix(J)
            # dx = np.linalg.solve(-J,dPQ)
            dx = scipy.sparse.linalg.spsolve(-J, dPQ)
            phi += dx[0 : int(dx.shape[0] / 2)]
            U_abs += dx[int(dx.shape[0] / 2) :]

            J = J.todense()

        u_all[j] = U
        iters_all[j] = iters
    return u_all, iters_all


# Z-Bus Jacobi
# ============
@cc.export(
    "zbusjacobi",
    "(complex128[:,:], complex128[:,:], complex128[:], complex64, float64, int64)",
)
def zbusjacobi(Zred, S_all, U0, slack_voltage, eps_s=1.0, max_iters=20):
    numberofnodes = S_all.shape[1]
    numberofloads = S_all.shape[0]
    iters_all = np.zeros(numberofloads, dtype=np.int32)
    U_all = np.zeros((numberofloads, numberofnodes + 1), dtype=np.complex128)
    U_all[:, 0] = slack_voltage
    U = np.ones(numberofnodes, dtype=np.complex128) * slack_voltage

    for i in range(numberofloads):
        s = S_all[i, :]
        iters = 0

        # In case of a voltage collapse in the previous computation, U must be reset
        # This is mostly jumped by branch prediction
        if np.abs(U)[0] == 0:
            U = np.ones(numberofnodes, dtype=np.complex128) * slack_voltage

        while True:
            Ibus = np.conj(s / U)
            U = Zred @ Ibus + slack_voltage
            iters += 1
            if np.max(np.abs(s - U * np.conj(Ibus))) < eps_s:
                break
            # Indicate voltage collapse with U == 0:
            if np.min(np.abs(U)) < 0 or iters > max_iters:
                U = np.zeros_like(U)
                iters = max_iters + 2
                break

        U_all[i, 1:] = U
        iters_all[i] = iters

    return U_all, iters_all


def zbusjacobi_sparse(Zred, S, slack_voltage, eps_s=1.0, max_iters=20):
    numberofnodes = S.shape[1]
    numberofloads = S.shape[0]
    iters_all = np.zeros(numberofloads, dtype=np.int32)
    U_all = np.zeros((numberofloads, numberofnodes), dtype=np.complex128)
    u = np.ones(numberofnodes, dtype=np.complex128) * slack_voltage
    Zred = scipy.sparse.csr_matrix(Zred)
    for i in range(numberofloads):
        s = S[i, :]
        iters = 0
        while True:
            iters += 1
            Ibus = np.conj(s / u)
            u = Zred.dot(Ibus) + slack_voltage
            if np.max(np.abs(s - u * np.conj(Ibus))) < eps_s or iters > max_iters:
                break
        U_all[i, :] = u
        iters_all[i] = iters

    return U_all, iters_all


@cc.export(
    "zbusjacobi_2slacks",
    "(complex128[:,:], complex128[:,:], complex128[:], complex128, int64, complex128, float64, float64, int64)",
)
def zbusjacobi_2slacks(
    Z,
    S_all,
    u0,
    slack_voltage,
    i_slack,
    U_slack2,
    eps_s=1.0,
    eps_u_slack=0.001,
    max_iters=20,
):
    numberofnodes = S_all.shape[1]
    numberofloads = S_all.shape[0]
    iters_all = np.zeros(numberofloads, dtype=np.int32)
    U_all = np.zeros((numberofloads, numberofnodes + 1), dtype=np.complex128)
    U_all[:, 0] = slack_voltage
    U = np.ones(numberofnodes, dtype=np.complex128) * slack_voltage

    for i in range(numberofloads):
        S = S_all[i, :]
        iters = 0
        while True:
            I = np.conj(S / U)
            I[0] = 0
            I[0] = (U_slack2 - slack_voltage - Z[0, :] @ I) / Z[0, 0]
            S[0] = U[0] * np.conj(I[0])
            U = Z @ I + slack_voltage
            iters += 1
            if (
                np.max(np.abs(S - U * np.conj(I))) < eps_s
                and np.abs(U[i_slack] - U_slack2) < eps_u_slack
            ) or iters > max_iters:
                break
        U_all[i, 1:] = U
        iters_all[i] = iters

    return U_all, iters_all


@cc.export(
    "zbusjacobi_mslacks",
    "(complex128[:,:], complex128[:,:], complex128[:], complex128, int64, complex128[:], float64, float64, int64)",
)
def zbusjacobi_mslacks(
    Zred,
    S_all,
    u0,
    slack_voltage,
    m,
    U_additional_slacks,
    eps_s=1.0,
    eps_u_slack=0.001,
    max_iters=20,
):
    """Perform the ZBus Jacobi Power Flow in a grid with n>2 slack nodes
    Zred    : [n-1 x n-1], complex128: Inverse of Yred, where the line and column of the main slack node have been deleted
    S_all   : [n-1 x nloads], complex128: Loads in W
    u0      : [n-1], complex128: Starting voltages
    slack_voltage : complex128: Main slack voltage
    m       : int64: number of additional slacks
    U_additional_slacks : [m], complex128: slack voltages of additional slacks
    eps_s   : float64: power residual as convergence condition
    eps_u_slack: float64: voltage residual as convergence condition for additional slacks
    max_iters: int64: number of iterations before assumed divergence and break

    Assumes Zred and S to be ordered, such that
    the nodes representing slack nodes are at the top
    """

    numberofnodes = S_all.shape[1]
    numberofloads = S_all.shape[0]
    iters_all = np.zeros(numberofloads, dtype=np.int32)
    U_all = np.zeros((numberofloads, numberofnodes + 1), dtype=np.complex128)
    U_all[:, 0] = slack_voltage
    U = np.ones(numberofnodes, dtype=np.complex128) * slack_voltage
    U_slack_vector = U.copy()

    for i in range(numberofloads):
        S = S_all[i, :]
        iters = 0
        while True:
            I = np.conj(S / U)
            I[:m] = np.linalg.solve(
                Zred[:m, :m],
                U_additional_slacks - U_slack_vector[:m] - Zred[:m, m:] @ I[m:],
            )
            U = Zred @ I + slack_voltage
            iters += 1
            S[:m] = U[:m] * np.conj(I[:m])
            if (
                np.max(np.abs(S - U * np.conj(I))) < eps_s
                and np.max(np.abs(U[:m] - U_additional_slacks)) < eps_u_slack
            ) or iters > max_iters:
                break
        U_all[i, 1:] = U
        iters_all[i] = iters

    return U_all, iters_all


@cc.export("zbusjacobi_1iter", "(complex128[:,:], complex128[:,:], complex64)")
def zbusjacobi_1iter(Zred, S_all, slack_voltage):
    numberofnodes = S_all.shape[1]
    numberofloads = S_all.shape[0]
    U_all = np.zeros((numberofloads, numberofnodes), dtype=np.complex128)
    U = np.ones(numberofnodes, dtype=np.complex128) * slack_voltage

    for i in range(numberofloads):
        s = S_all[i, :]
        U = Zred @ np.conj(s / U) + slack_voltage
        U_all[i, :] = U

    return U_all


@cc.export("zbusjacobi_2iter", "(complex128[:,:], complex128[:,:], complex64)")
def zbusjacobi_1iter(Zred, S_all, slack_voltage):
    numberofnodes = S_all.shape[1]
    numberofloads = S_all.shape[0]
    U_all = np.zeros((numberofloads, numberofnodes), dtype=np.complex128)
    U = np.ones(numberofnodes, dtype=np.complex128) * slack_voltage

    for i in range(numberofloads):
        s = S_all[i, :]
        U = Zred @ np.conj(s / U) + slack_voltage
        U = Zred @ np.conj(s / U) + slack_voltage
        U_all[i, :] = U

    return U_all


# @cc.export('zbusjacobi','(complex128[:,:], complex128[:,:], complex64, float64, int64)')
# def zbusjacobi(Zred, S_all, slack_voltage, eps_s=1.0, max_iters=20):
#     numberofnodes = S_all.shape[1]
#     numberofloads = S_all.shape[0]
#     iters_all = np.zeros(numberofloads, dtype=np.int32)
#     U_all = np.zeros((numberofloads, numberofnodes), dtype=np.complex128)
#     U = np.ones(numberofnodes, dtype=np.complex128) * slack_voltage

#     for i in range(numberofloads):
#         s = S_all[i,:]
#         iters = 0

#         if np.abs(U)[0] == 0:
#             U = np.ones(numberofnodes, dtype=np.complex128) * slack_voltage

#         while True:
#             Ibus = np.conj(s/U)
#             U = Zred @ Ibus + slack_voltage
#             iters += 1
#             if np.max(np.abs(s - U * np.conj(Ibus))) < eps_s:
#                 break
#             if np.min(np.abs(U)) < 0 or iters > max_iters:
#                 U = np.zeros_like(U)
#                 break
#         U_all[i,:] = U
#         iters_all[i] = iters

#     return U_all, iters_all


# Backward/Forward Sweep for Feeders
# ==================================
@cc.export("bfs", "(complex128[:], complex128[:,:], complex64, float64, int64)")
def bfs(Zline, S_all, slack_voltage, eps_s=1.0, max_iters=100):
    S_all = S_all[:, 1:]
    Ubus = np.ones(S_all.shape[1], dtype=np.complex128) * slack_voltage
    Ibus = np.zeros(S_all.shape[1], dtype=np.complex128)
    U_all = np.zeros((S_all.shape[0], Ubus.shape[0]), dtype=np.complex128)
    iters_all = np.zeros(S_all.shape[0], dtype=np.int32)

    for i in range(S_all.shape[0]):
        s = S_all[i, :]
        iters = 0
        while True:
            Ibus = np.conj(s / Ubus)
            Ubus = slack_voltage - np.cumsum(np.cumsum(Ibus[::-1])[::-1] * Zline)
            iters += 1
            if np.max(np.abs(s - Ubus * np.conj(Ibus))) < eps_s or iters > max_iters:
                break
        U_all[i, :] = Ubus
        iters_all[i] = iters

    return U_all, iters_all


def compile(verbose=False):
    cc.verbose = verbose
    cc.compile()


if __name__ == "__main__":
    compile()
